import torch
import numpy as np

from tqdm import tqdm

def get_edges_from_faces(faces):
    """
    Extract unique edges from faces.
    """
    edges = torch.cat([
        faces[:, [0, 1]],
        faces[:, [1, 2]],
        faces[:, [2, 0]]
    ], dim=0)
    edges = torch.sort(edges, dim=1).values  # Sort to ensure uniqueness
    edges = torch.unique(edges, dim=0)
    return edges

def runBatchedSMPLX(params, smplx_model, BS=500, with_vertices=False, device='cuda'):
    params = {k: v.to(device) for k, v in params.items()}
    out = {'joints': []}
    if with_vertices:
        out['vertices'] = []
    num_data = 1
    for key in params.keys():
        num_data = max(num_data, params[key].shape[0])
    if BS > num_data:
        out_ = smplx_model(**params)
        out['joints'] = out_.joints.detach().cpu()
        if with_vertices:
            out['vertices'] = out_.vertices.detach().cpu()
    else:
        for i in tqdm(range(0, num_data, BS)):
            batch_params = {k: v if v.shape[0] == 1 else v[i:i+BS] for k, v in params.items()}
            if 'J_regressor' in params:
                batch_params['J_regressor'] = params['J_regressor']
            if 'lbs_weights' in params:
                batch_params['lbs_weights'] = params['lbs_weights']
            out_ = smplx_model(**batch_params)
            if with_vertices:
                out['vertices'] += [out_['vertices'].detach().cpu()]
            out['joints'] += [out_['joints'].detach().cpu()]

        for key in out.keys():
            out[key] = torch.cat(out[key], 0)
    return out

def getModelOut(model, params, BS=512, with_vertices=True):
    with torch.no_grad():
        out = runBatchedSMPLX(params, model, BS=BS, with_vertices=with_vertices)
    return out

def getSMPLXOut(smplx_kwargs, device, params, BS=512, run_with_verts=True):
    from .smpl_fx.body_models import SMPLFX
    model = SMPLFX(**smplx_kwargs).to(device)
    out = runBatchedSMPLX(params, model, BS=BS, with_vertices=run_with_verts, device=device)
    return out, model

def compute_normals_diff(vertices, faces, eps=1e-6):
    """
    Differentiable computation of vertex and face normals.
    Args:
        vertices: (V, 3)
        faces: (F, 3) long
    Returns:
        vertex_normals: (V, 3)
        face_normals: (F, 3)
    """
    v0 = vertices[faces[:, 0]]  # (F, 3)
    v1 = vertices[faces[:, 1]]
    v2 = vertices[faces[:, 2]]

    face_normals = torch.cross(v1 - v0, v2 - v0, dim=1)  # (F, 3)
    face_normals = face_normals / (face_normals.norm(dim=1, keepdim=True) + eps)

    # Prepare indices to scatter into vertex normals
    F = faces.shape[0]
    V = vertices.shape[0]

    idx = faces.reshape(-1)          # (3*F,)
    face_normals_repeat = face_normals.repeat_interleave(3, dim=0)  # (3*F, 3)

    vertex_normals = torch.zeros_like(vertices)
    vertex_normals = vertex_normals.scatter_add(0, idx[:, None].expand(-1, 3), face_normals_repeat)
    vertex_normals = vertex_normals / (vertex_normals.norm(dim=1, keepdim=True) + eps)

    return vertex_normals, face_normals

def convertSMPLX2SMPLFXKwargs(kwargs, num_betas_flame=100, v_mask_path_flame=None, corr_fname=None):
	kwargs['gender'] = 'neutral'
	kwargs.setdefault('ext', 'pkl')
	if v_mask_path_flame is None:
		raise ValueError("v_mask_path_flame must be provided for SMPLFX model")
	kwargs['v_mask_path_flame'] = v_mask_path_flame
	if corr_fname is None:
		raise ValueError("corr_fname must be provided for SMPLFX model")
	kwargs['corr_fname'] = corr_fname
	kwargs['use_eyelids'] = True
	kwargs['add_teeth'] = True
	kwargs['num_betas_flame'] = num_betas_flame
	kwargs['with_mouth_interior'] = True
	kwargs['mount_interior_resolution_u'] = 10
	kwargs['mount_interior_resolution_v'] = 20
	return kwargs


def flipPoseRot(poses):
    # poses ... x 3 x 3
    assert poses.shape[-1] == 3 and poses.shape[-2] == 3
    flip_mask = torch.tensor([
        [ 1, -1, -1],
        [-1,  1,  1],
        [-1,  1,  1]
    ], device=poses.device, dtype=poses.dtype)
    while flip_mask.ndim < poses.ndim:
        flip_mask = flip_mask[None]
    flipped_poses = poses * flip_mask
    return flipped_poses


import pytorch3d
def remapIndependentHandToBody(right_hand_orient, left_hand_orient, body_pose, global_orient, hands_format, body_format, conv='ZYX'):
    """
    Given hand poses extracted separately (e.g., HaMeR/WiLor) this code remaps their more accurate hand orientation to the body pose.
    """
    assert len(right_hand_orient) == len(left_hand_orient) == len(body_pose) == len(global_orient)

    if body_format == 'euler':
        assert body_pose.ndim == 2
        assert global_orient.ndim == 2
        body_pose_mat = pytorch3d.transforms.euler_angles_to_matrix(body_pose.reshape(-1, 21, 3), conv)
        global_orient_mat = pytorch3d.transforms.euler_angles_to_matrix(global_orient, conv)
    elif body_format == 'rot':
        assert body_pose.ndim == 4
        assert global_orient.ndim == 3
        body_pose_mat = body_pose
        global_orient_mat = global_orient
    else:
        raise NotImplementedError
    
    R_global_elbow_l = global_orient_mat @ body_pose_mat[:,2] @ body_pose_mat[:,5] @ body_pose_mat[:,8] @ body_pose_mat[:,11] @ body_pose_mat[:,15] @ body_pose_mat[:,17]
    R_global_elbow_r = global_orient_mat @ body_pose_mat[:,2] @ body_pose_mat[:,5] @ body_pose_mat[:,8] @ body_pose_mat[:,11] @ body_pose_mat[:,16] @ body_pose_mat[:,18]

    if hands_format == 'euler':
        assert right_hand_orient.ndim == 2
        assert left_hand_orient.ndim == 2
        right_hand_orient_mat = pytorch3d.transforms.euler_angles_to_matrix(right_hand_orient, conv)
        left_hand_orient_mat = pytorch3d.transforms.euler_angles_to_matrix(left_hand_orient, conv)
    elif hands_format == 'rot':
        assert right_hand_orient.ndim == 3
        assert left_hand_orient.ndim == 3
        right_hand_orient_mat = right_hand_orient
        left_hand_orient_mat = left_hand_orient
    else:
        raise NotImplementedError

    body_elbow_r = R_global_elbow_r.transpose(-1,-2) @ right_hand_orient_mat
    body_elbow_l = R_global_elbow_l.transpose(-1,-2) @ left_hand_orient_mat

    body_pose = body_pose.clone()
    if body_format == 'euler':
        body_elbow_r = pytorch3d.transforms.matrix_to_euler_angles(body_elbow_r, conv)
        body_elbow_l = pytorch3d.transforms.matrix_to_euler_angles(body_elbow_l, conv)
        body_pose[...,19*3:19*3+3] = body_elbow_l # left hand
        body_pose[...,20*3:20*3+3] = body_elbow_r # right hand
    elif body_format == 'rot':
        body_pose[...,19,:,:] = body_elbow_l
        body_pose[...,20,:,:] = body_elbow_r

    return body_pose


def cross_product_indices(idxs1, idxs2):
    all_indices = []
    for i in range(len(idxs1)):
        for j in range(len(idxs2)):
            all_indices.append([idxs1[i], idxs2[j]])
    return np.array(all_indices)

class Contacts():
    def __init__(self, on_joints=True, is_1to1_connections=True):
        self.on_joints = on_joints
        self.is_1to1_connections = is_1to1_connections
        self.right_hand_joints = np.array([21, 52, 53, 54, 71, 40, 41, 42, 72, 43, 44, 45, 73, 49, 50, 51, 74, 46, 47, 48, 75])
        self.left_hand_joints = np.array([20, 37, 38, 39, 66, 25, 26, 27, 67, 28, 29, 30, 68, 34, 35, 36, 69, 31, 32, 33, 70])
        self.face_indices = np.array(np.arange(95, 107, 1).tolist() + np.arange(107, 119, 1).tolist())
        if on_joints:
            if is_1to1_connections:
                self.contact_connections = np.concatenate((self.right_hand_joints[:,None], self.left_hand_joints[:,None]), axis=-1)
            else:
                self.contact_connections = cross_product_indices(self.right_hand_joints, self.left_hand_joints)

            self.contact_connections = np.concatenate((self.contact_connections, cross_product_indices(self.right_hand_joints, self.face_indices)), axis=0)
            self.contact_connections = np.concatenate((self.contact_connections, cross_product_indices(self.left_hand_joints, self.face_indices)), axis=0)            
        else:
            raise NotImplementedError()

    def get_relative_contacts_joints(self, joints):
        return joints[:, self.contact_connections[:,1]] - joints[:, self.contact_connections[:,0]]

    def get_relative_contacts(self, output):
        if self.on_joints:
            return self.get_relative_contacts_joints(output.joints)
        else:
            raise NotImplementedError()

    

def getContactWeights(relative_contacts, max_dist_contact):
    contact_distances = torch.linalg.norm(relative_contacts, dim=-1, keepdim=True)
    mask = (contact_distances < max_dist_contact).float()
    return mask



class SMPLX2DJointMapper():
	def __init__(self, with_hips=False, with_legs=True, with_top_feet=True, with_mp_main_face=False, with_hand_palm=False, with_shoulders=True, detector_name='rtmp', use_face_contour=False, is_full_face=False, use_full_mouth=False):
		if use_face_contour:
			assert is_full_face, "use_face_contour can be True only if is_full_face is also True"
		self.use_face_contour = use_face_contour
		if with_shoulders:
			smplx_body_idxs = [17, 19, 21, 16, 18, 20]
			self.edges = [[0,1], [1,2], [3,4], [4,5], [0,3]]

			if detector_name == 'rtmp':
				body_idxs = [6, 8, 10, 5, 7, 9]
			elif detector_name == 'openpose-25':
				body_idxs = [2, 3, 4, 5, 6, 7]
			else: raise ValueError('Unknown detector name: ' + detector_name)
		else:
			smplx_body_idxs = [19, 21, 18, 20]
			self.edges = [[0,1], [2,3]]

			if detector_name == 'rtmp':
				body_idxs = [8, 10, 7, 9]
			elif detector_name == 'openpose-25':
				body_idxs = [3, 4, 6, 7]
			else: raise ValueError('Unknown detector name: ' + detector_name)

		if with_hips:
			smplx_body_idxs += [2, 1]	# right hip, left hip
			self.edges += [[len(smplx_body_idxs)-2, len(smplx_body_idxs)-1]]

			if detector_name == 'rtmp':
				body_idxs += [12, 11] # right hip, left hip
			elif detector_name == 'openpose-25':
				body_idxs += [9, 12]
			else: raise ValueError('Unknown detector name: ' + detector_name)

		if with_legs:
			smplx_body_idxs += [5, 8, 4, 7]
			self.edges += [[len(smplx_body_idxs)-4, len(smplx_body_idxs)-3], [len(smplx_body_idxs)-2, len(smplx_body_idxs)-1]]
			if detector_name == 'rtmp':
				body_idxs += [14, 16, 13, 15]
			elif detector_name == 'openpose-25':
				assert False
			else: raise ValueError('Unknown detector name: ' + detector_name)

			if with_top_feet:
				smplx_body_idxs += [63, 60]

				if detector_name == 'rtmp':
					body_idxs += [20, 17]
				elif detector_name == 'openpose-25':
					assert False
				else: raise ValueError('Unknown detector name: ' + detector_name)

		if with_mp_main_face:
			smplx_body_idxs += [95, 104] + [107, 113]

			if detector_name == 'rtmp':
				body_idxs += [59, 68] + [71, 77]
			elif detector_name == 'openpose-25':
				assert False
			else: raise ValueError('Unknown detector name: ' + detector_name)

		if with_hand_palm:
			npts = len(smplx_body_idxs)
			smplx_body_idxs += [40, 46] + [25, 31]
			self.edges += [[2, npts  ], [2, npts+1], [npts  , npts+1]]
			self.edges += [[5, npts+2], [5, npts+3], [npts+2, npts+3]]

			if detector_name == 'rtmp':
				body_idxs += [117, 129] + [96, 108]
			elif detector_name == 'openpose-25':
				assert False
			else: raise ValueError('Unknown detector name: ' + detector_name)

		self.smplx_rhand_idxs = [21, 52, 53, 54, 71, 40, 41, 42, 72, 43, 44, 45, 73, 49, 50, 51, 74, 46, 47, 48, 75]
		self.smplx_lhand_idxs = [20, 37, 38, 39, 66, 25, 26, 27, 67, 28, 29, 30, 68, 34, 35, 36, 69, 31, 32, 33, 70]

		# smplx_face_idxs_no_contour = [76,77,78,79,80] + [81,82,83,84,85] + [86,87,88,89] + [90,91,92,93,94] + [95,96,97,98,99,100] + [104,103,102,101,106,105] + [107,108,109,110,111,112,113,114,115,116,117,118] + [119,120,121,122,123,124,125,126] + [24, 23]

		smplx_face_idxs_no_contour = np.arange(76, 127).tolist() + [24, 23]

		npts = len(body_idxs) + len(self.smplx_rhand_idxs) + len(self.smplx_lhand_idxs)
		self.face_edges = []
		if detector_name == 'rtmp':
			self.rhand_idxs = np.arange(133-21, 133).tolist()
			self.lhand_idxs = np.arange(133-42, 133-21).tolist()
			if is_full_face:
				if use_face_contour:
					self.face_idxs = np.arange(23, 23+68).tolist() + [2, 1]
					self.smplx_face_idxs = np.arange(127, 127+17).tolist() + smplx_face_idxs_no_contour
					self.mouth_face_subidxs = None
				else:
					self.face_idxs = np.arange(40, 23+68).tolist() + [2, 1]
					self.smplx_face_idxs = smplx_face_idxs_no_contour
					self.mouth_face_subidxs = None
			else:
				if use_full_mouth:
					self.face_idxs = (np.arange(59, 71).tolist() + np.arange(71, 91, 1).tolist())
					self.smplx_face_idxs = np.arange(95, 107, 1).tolist() + np.arange(107, 127, 1).tolist()
					self.face_edges += getLoopedEdgesPairs(6) + getLoopedEdgesPairs(6,6) + getLoopedEdgesPairs(12,12) + getLoopedEdgesPairs(8,24)
					self.mouth_face_subidxs = np.arange(len(self.face_idxs)-20, len(self.face_idxs), 1)
				else:
					self.face_idxs = (np.arange(59, 71).tolist() + np.arange(71, 91, 1).tolist())[:24]
					self.smplx_face_idxs = np.arange(95, 107, 1).tolist() + np.arange(107, 119, 1).tolist()
					self.mouth_face_subidxs = np.arange(len(self.face_idxs)-12, len(self.face_idxs), 1)
	
		elif detector_name == 'openpose-25':
			self.rhand_idxs = np.arange(25, 25+21).tolist()
			self.lhand_idxs = np.arange(25+21, 25+42).tolist()
			if is_full_face:
				# right eyebrow, left eyebrow, nose bridge, nose bottom, right eye, left eye, outer lips, inner lips, pupils
				if use_face_contour:
					self.face_idxs = np.arange(25+42, 25+42+70).tolist()
					self.smplx_face_idxs = np.arange(127, 127+17).tolist() + smplx_face_idxs_no_contour
					# openpose scheme: https://www.researchgate.net/figure/Openpose-detected-body-hand-and-face-keypoints_fig2_353668783
					self.mouth_face_subidxs = None
				else:
					self.face_idxs = np.arange(25+42+17, 25+42+70).tolist()
					self.smplx_face_idxs = smplx_face_idxs_no_contour
					self.mouth_face_subidxs = None
			else:
				if use_full_mouth:
					self.face_idxs = (np.arange(25+42, 25+42+12).tolist() + np.arange(25+42+12, 25+42+32, 1).tolist())
					self.smplx_face_idxs = np.arange(95, 107, 1).tolist() + np.arange(107, 127, 1).tolist()
					self.mouth_face_subidxs = np.arange(len(self.face_idxs)-20, len(self.face_idxs), 1)
				else:
					self.face_idxs = (np.arange(25+42, 25+42+12).tolist() + np.arange(25+42+12, 25+42+12+12, 1).tolist())
					self.smplx_face_idxs = np.arange(95, 107, 1).tolist() + np.arange(107, 119, 1).tolist()
					self.mouth_face_subidxs = np.arange(len(self.face_idxs)-12, len(self.face_idxs), 1)
		else:
			raise ValueError('Unknown detector name: ' + detector_name)

		if len(self.face_edges) > 0:
			self.edges = self.edges + (npts + np.array(self.face_edges)).tolist()

		self.smplx_hand_idxs = self.smplx_rhand_idxs + self.smplx_lhand_idxs
		self.hand_idxs = self.rhand_idxs + self.lhand_idxs

		edges_one_hand = [[0, 1], [1, 2], [2, 3], [3, 4], [0, 5], [5, 6], [6, 7], [7, 8], [5, 9], [9, 10], [10, 11], [11, 12], [9, 13], [13, 14], [14, 15], [15, 16], [13, 17], [17, 18], [18, 19], [19, 20], [17, 0]]
		self.edges += (len(smplx_body_idxs) + np.array(edges_one_hand)).tolist() + (len(smplx_body_idxs) + len(self.smplx_rhand_idxs) + np.array(edges_one_hand)).tolist()
		self.smplx_idxs = smplx_body_idxs + self.smplx_rhand_idxs + self.smplx_lhand_idxs + self.smplx_face_idxs

		self.body_idxs = body_idxs
		self.smplx_body_idxs = smplx_body_idxs

		self.idxs = body_idxs + self.rhand_idxs + self.lhand_idxs + self.face_idxs
		self.face_subidxs = np.arange(len(self.idxs)-len(self.face_idxs), len(self.idxs), 1) # same for 2D detections and SMPLX
		self.body_subidxs = np.arange(0, len(self.body_idxs), 1)
		self.hand_subidxs = np.arange(len(self.body_idxs), len(self.body_idxs)+len(self.hand_idxs), 1)

		self.smplx_idxs_no_face = self.smplx_idxs[:-len(self.smplx_face_idxs)]
		self.idxs_no_face = self.idxs[:-len(self.face_idxs)]

		self.smplx_body_mask_no_face = torch.zeros(len(self.smplx_idxs) - len(self.smplx_face_idxs), dtype=torch.bool)
		self.smplx_body_mask_no_face[:len(self.smplx_body_idxs)] = True
		self.smplx_hand_mask_no_face = torch.zeros(len(self.smplx_idxs) - len(self.smplx_face_idxs), dtype=torch.bool)
		self.smplx_hand_mask_no_face[len(self.smplx_body_idxs):] = True

		assert len(self.idxs) == len(self.smplx_idxs)
		assert np.array(self.edges).max() < len(self.idxs)



"""Head movement detectors: nods, shakes, and tilts."""

import numpy as np
from scipy.ndimage import gaussian_filter1d, median_filter
from scipy.signal import find_peaks

from ..constants import FPS
from ..models import NMSEvent
from ..processing.utils import build_idx_to_frame, mask_to_regions, merge_regions


def detect_head_oscillations(
    signal,
    successful_frames,
    valid,
    segments,
    *,
    tier="HEAD-NOD",
    sigma=1.0,
    min_amplitude=0.03,
    window_frames=8,
    min_crossings=2,
    merge_gap=4,
    min_duration_frames=2,
    idx_to_frame=None,
):
    """Detect oscillatory head movements (nods or shakes) via zero-crossing analysis."""
    if idx_to_frame is None:
        idx_to_frame = build_idx_to_frame(successful_frames)
    events = []

    for seg_start, seg_end in segments:
        seg = signal[seg_start:seg_end].astype(np.float64)
        if len(seg) < window_frames + 2:
            continue

        smoothed = gaussian_filter1d(seg, sigma=sigma)
        velocity = np.diff(smoothed)

        signs = np.sign(velocity)
        crossings = np.where(np.diff(signs) != 0)[0]

        if len(crossings) < min_crossings:
            continue

        osc_mask = np.zeros(len(seg), dtype=bool)
        left_ptr = 0
        for i in range(len(seg) - window_frames):
            while left_ptr < len(crossings) and crossings[left_ptr] < i:
                left_ptr += 1
            right_ptr = left_ptr
            while right_ptr < len(crossings) and crossings[right_ptr] < i + window_frames:
                right_ptr += 1
            window_crossings = right_ptr - left_ptr
            if window_crossings >= min_crossings:
                window_sig = smoothed[i : i + window_frames]
                amplitude = window_sig.max() - window_sig.min()
                if amplitude >= min_amplitude:
                    osc_mask[i : i + window_frames] = True

        regions = mask_to_regions(osc_mask)
        regions = merge_regions(regions, merge_gap)

        for start, end in regions:
            duration = end - start
            if duration < min_duration_frames:
                continue

            seg_slice = smoothed[start:end]
            peak_idx = start + int(np.argmax(np.abs(seg_slice - np.mean(seg_slice))))
            peak_val = float(seg_slice.max() - seg_slice.min())

            global_start = seg_start + start
            global_end = seg_start + end - 1
            global_peak = seg_start + peak_idx

            region_crossings = np.sum((crossings >= start) & (crossings < end))
            crossing_density = region_crossings / (duration / window_frames)
            amp_score = float(np.clip((peak_val - min_amplitude) / (3 * min_amplitude), 0, 1))
            density_score = float(np.clip((crossing_density - min_crossings) / min_crossings, 0, 1))
            confidence = 0.5 * amp_score + 0.5 * density_score

            events.append(
                NMSEvent(
                    tier=tier,
                    start_frame=idx_to_frame.get(global_start, global_start),
                    end_frame=idx_to_frame.get(global_end, global_end),
                    peak_frame=idx_to_frame.get(global_peak, global_peak),
                    peak_value=peak_val,
                    duration_ms=duration * (1000.0 / FPS),
                    confidence=float(np.clip(confidence, 0, 1)),
                    label=tier.lower().replace("-", " "),
                )
            )

    return events


def detect_single_head_movements(
    signal,
    successful_frames,
    valid,
    segments,
    *,
    tier="HEAD-NOD",
    sigma=1.0,
    min_amplitude=0.01,
    existing_events=None,
    idx_to_frame=None,
):
    """Detect isolated single head movements via velocity peaks."""
    if idx_to_frame is None:
        idx_to_frame = build_idx_to_frame(successful_frames)

    covered = np.zeros(len(signal), dtype=bool)
    if existing_events:
        frame_to_idx = {}
        for i, vf in enumerate(successful_frames):
            frame_to_idx[int(vf)] = i
        for evt in existing_events:
            if evt.tier == tier:
                for f in range(evt.start_frame, evt.end_frame + 1):
                    idx = frame_to_idx.get(f)
                    if idx is not None:
                        covered[idx] = True

    events = []

    for seg_start, seg_end in segments:
        seg = signal[seg_start:seg_end].astype(np.float64)
        if len(seg) < 4:
            continue

        smoothed = gaussian_filter1d(seg, sigma=sigma)
        velocity = np.diff(smoothed)

        abs_vel = np.abs(velocity)
        peaks, props = find_peaks(abs_vel, prominence=min_amplitude * 0.4, distance=3)

        if len(peaks) == 0:
            continue

        for pk in peaks:
            global_pk = seg_start + pk
            if covered[global_pk]:
                continue

            left = pk
            while left > 0 and np.sign(velocity[left - 1]) == np.sign(velocity[pk]):
                left -= 1
            right = pk + 1
            while right < len(velocity) and np.sign(velocity[right]) == np.sign(velocity[pk]):
                right += 1

            amplitude = abs(smoothed[right] - smoothed[left]) if right < len(smoothed) else 0
            if amplitude < min_amplitude:
                continue

            start = max(0, left - 1)
            end = min(len(seg), right + 1)
            duration = end - start
            if duration < 2:
                continue

            global_start = seg_start + start
            global_end = seg_start + end - 1

            region_covered = covered[global_start : global_end + 1].mean()
            if region_covered > 0.5:
                continue

            peak_val = float(amplitude)
            amp_score = float(np.clip((amplitude - min_amplitude) / (3 * min_amplitude), 0, 1))
            confidence = 0.7 * amp_score + 0.3 * 0.5

            label_prefix = tier.lower().replace("-", " ")
            events.append(
                NMSEvent(
                    tier=tier,
                    start_frame=idx_to_frame.get(global_start, global_start),
                    end_frame=idx_to_frame.get(global_end, global_end),
                    peak_frame=idx_to_frame.get(global_pk, global_pk),
                    peak_value=peak_val,
                    duration_ms=duration * (1000.0 / FPS),
                    confidence=float(np.clip(confidence, 0, 1)),
                    label=f"single {label_prefix}",
                )
            )

    return events


def detect_head_tilts(
    signal,
    successful_frames,
    valid,
    segments,
    *,
    sigma=3.0,
    threshold=0.06,
    baseline_window=251,
    min_duration_frames=2,
    idx_to_frame=None,
):
    """Detect sustained head tilts from roll signal."""
    if idx_to_frame is None:
        idx_to_frame = build_idx_to_frame(successful_frames)
    events = []

    for seg_start, seg_end in segments:
        seg = signal[seg_start:seg_end].astype(np.float64)
        if len(seg) < min_duration_frames + 2:
            continue

        smoothed = gaussian_filter1d(seg, sigma=sigma)
        baseline = median_filter(smoothed, size=baseline_window, mode="reflect")

        deviation = smoothed - baseline
        tilt_mask = np.abs(deviation) > threshold

        regions = mask_to_regions(tilt_mask)

        for start, end in regions:
            duration = end - start
            if duration < min_duration_frames:
                continue

            seg_dev = deviation[start:end]
            peak_idx = start + int(np.argmax(np.abs(seg_dev)))
            peak_val = float(deviation[peak_idx])

            global_start = seg_start + start
            global_end = seg_start + end - 1
            global_peak = seg_start + peak_idx

            mean_dev = float(np.mean(seg_dev))
            direction = "tilt-left" if mean_dev > 0 else "tilt-right"

            amp_score = float(np.clip((abs(peak_val) - threshold) / (3 * threshold), 0, 1))
            dur_score = float(np.clip(duration / (FPS * 0.5), 0, 1))
            confidence = 0.6 * amp_score + 0.4 * dur_score

            events.append(
                NMSEvent(
                    tier="HEAD-TILT",
                    start_frame=idx_to_frame.get(global_start, global_start),
                    end_frame=idx_to_frame.get(global_end, global_end),
                    peak_frame=idx_to_frame.get(global_peak, global_peak),
                    peak_value=peak_val,
                    duration_ms=duration * (1000.0 / FPS),
                    confidence=float(np.clip(confidence, 0, 1)),
                    label=direction,
                )
            )

    return events



"""Tests for research utilities."""

import json

import numpy as np
import pytest

from sltk.data.core import Segment, SegmentList
from sltk.exceptions import ConfigurationError
from sltk.utils.research import (
    DatasetStats,
    compare_vocabularies,
    compute_pose_coverage,
    create_vocabulary_mapping,
    export_segments_for_training,
    find_gloss_examples,
)


class TestDatasetStats:
    """Tests for DatasetStats dataclass."""

    def test_creation(self):
        """Test basic DatasetStats creation."""
        stats = DatasetStats(
            name="test",
            num_samples=10,
            vocabulary_size=100,
        )
        assert stats.name == "test"
        assert stats.num_samples == 10
        assert stats.vocabulary_size == 100

    def test_to_dict(self):
        """Test conversion to dictionary."""
        stats = DatasetStats(
            name="test",
            num_samples=10,
            gloss_frequencies={"HELLO": 50},
        )
        d = stats.to_dict()
        assert d["name"] == "test"
        assert d["num_samples"] == 10
        assert d["gloss_frequencies"]["HELLO"] == 50

    def test_json_roundtrip(self, tmp_path):
        """Test JSON save and load."""
        stats = DatasetStats(
            name="test",
            num_samples=10,
            vocabulary_size=100,
            gloss_frequencies={"HELLO": 50, "WORLD": 30},
        )

        json_path = tmp_path / "stats.json"
        stats.to_json(json_path)

        loaded = DatasetStats.from_json(json_path)
        assert loaded.name == stats.name
        assert loaded.num_samples == stats.num_samples
        assert loaded.vocabulary_size == stats.vocabulary_size
        assert loaded.gloss_frequencies == stats.gloss_frequencies


class TestCompareVocabularies:
    """Tests for vocabulary comparison."""

    def test_identical_vocabularies(self):
        """Test comparison of identical vocabularies."""
        stats1 = DatasetStats(
            name="A",
            gloss_frequencies={"HELLO": 50, "WORLD": 30},
        )
        stats2 = DatasetStats(
            name="B",
            gloss_frequencies={"HELLO": 40, "WORLD": 35},
        )

        result = compare_vocabularies(stats1, stats2)

        assert result["intersection_size"] == 2
        assert result["jaccard_index"] == 1.0
        assert result["overlap_ratio_1"] == 1.0
        assert result["overlap_ratio_2"] == 1.0

    def test_disjoint_vocabularies(self):
        """Test comparison of disjoint vocabularies."""
        stats1 = DatasetStats(
            name="A",
            gloss_frequencies={"HELLO": 50, "WORLD": 30},
        )
        stats2 = DatasetStats(
            name="B",
            gloss_frequencies={"GOODBYE": 40, "THANKS": 35},
        )

        result = compare_vocabularies(stats1, stats2)

        assert result["intersection_size"] == 0
        assert result["jaccard_index"] == 0.0
        assert result["only_in_1_count"] == 2
        assert result["only_in_2_count"] == 2

    def test_partial_overlap(self):
        """Test comparison with partial overlap."""
        stats1 = DatasetStats(
            name="A",
            gloss_frequencies={"HELLO": 50, "WORLD": 30, "FOO": 10},
        )
        stats2 = DatasetStats(
            name="B",
            gloss_frequencies={"HELLO": 40, "BAR": 35},
        )

        result = compare_vocabularies(stats1, stats2)

        assert result["intersection_size"] == 1
        assert result["vocab_size_1"] == 3
        assert result["vocab_size_2"] == 2
        assert "HELLO" in result["common_glosses"]


class TestCreateVocabularyMapping:
    """Tests for vocabulary mapping creation."""

    def test_with_elan_files(self, tmp_path):
        """Test vocabulary mapping from ELAN files."""
        from sltk.io.elan import write_eaf

        # Create test ELAN files
        segments1 = SegmentList(
            [
                Segment(0, 1, "HELLO"),
                Segment(1, 2, "WORLD"),
                Segment(2, 3, "HELLO"),  # Duplicate
            ]
        )
        segments2 = SegmentList(
            [
                Segment(0, 1, "HELLO"),
                Segment(1, 2, "FOO"),
            ]
        )

        eaf1 = tmp_path / "test1.eaf"
        eaf2 = tmp_path / "test2.eaf"
        write_eaf(segments1, eaf1)
        write_eaf(segments2, eaf2)

        mapping = create_vocabulary_mapping(
            elan_files=[eaf1, eaf2],
            min_frequency=1,
            special_tokens=["<PAD>", "<UNK>"],
        )

        assert mapping["vocab_size"] >= 5  # 2 special + 3 glosses
        assert "<PAD>" in mapping["gloss_to_id"]
        assert "<UNK>" in mapping["gloss_to_id"]
        assert "HELLO" in mapping["gloss_to_id"]
        assert mapping["frequencies"]["HELLO"] == 3

    def test_min_frequency_filter(self, tmp_path):
        """Test minimum frequency filtering."""
        from sltk.io.elan import write_eaf

        # Create test file with varying frequencies
        segments = SegmentList(
            [
                Segment(0, 1, "COMMON"),
                Segment(1, 2, "COMMON"),
                Segment(2, 3, "COMMON"),
                Segment(3, 4, "RARE"),  # Only once
            ]
        )

        eaf = tmp_path / "test.eaf"
        write_eaf(segments, eaf)

        mapping = create_vocabulary_mapping(
            elan_files=[eaf],
            min_frequency=2,  # Filter out RARE
            special_tokens=None,
        )

        assert "COMMON" in mapping["gloss_to_id"]
        assert "RARE" not in mapping["gloss_to_id"]


class TestExportSegmentsForTraining:
    """Tests for training data export."""

    def test_csv_export(self, tmp_path):
        """Test CSV export."""
        import csv

        from sltk.io.elan import write_eaf

        segments = SegmentList(
            [
                Segment(0, 1, "HELLO"),
                Segment(1, 2, "WORLD"),
            ]
        )

        eaf = tmp_path / "test.eaf"
        write_eaf(segments, eaf)

        output = tmp_path / "output.csv"
        export_segments_for_training(
            elan_files=[eaf],
            output_path=output,
            format="csv",
            include_context=False,
        )

        assert output.exists()
        with open(output) as f:
            reader = csv.DictReader(f)
            rows = list(reader)

        assert len(rows) == 2
        assert rows[0]["label"] == "HELLO"
        assert rows[1]["label"] == "WORLD"

    def test_json_export(self, tmp_path):
        """Test JSON export."""
        from sltk.io.elan import write_eaf

        segments = SegmentList(
            [
                Segment(0, 1, "HELLO"),
            ]
        )

        eaf = tmp_path / "test.eaf"
        write_eaf(segments, eaf)

        output = tmp_path / "output.json"
        export_segments_for_training(
            elan_files=[eaf],
            output_path=output,
            format="json",
        )

        with open(output) as f:
            data = json.load(f)

        assert len(data) == 1
        assert data[0]["label"] == "HELLO"

    def test_context_inclusion(self, tmp_path):
        """Test context window inclusion."""
        import csv

        from sltk.io.elan import write_eaf

        segments = SegmentList(
            [
                Segment(0, 1, "A"),
                Segment(1, 2, "B"),
                Segment(2, 3, "C"),
            ]
        )

        eaf = tmp_path / "test.eaf"
        write_eaf(segments, eaf)

        output = tmp_path / "output.csv"
        export_segments_for_training(
            elan_files=[eaf],
            output_path=output,
            format="csv",
            include_context=True,
            context_window=1,
        )

        with open(output) as f:
            reader = csv.DictReader(f)
            rows = list(reader)

        # Middle segment should have context
        middle = rows[1]
        assert middle["label"] == "B"
        assert "A" in middle["context_before"]
        assert "C" in middle["context_after"]


class TestFindGlossExamples:
    """Tests for gloss example search."""

    def test_find_examples(self, tmp_path):
        """Test finding gloss examples."""
        from sltk.io.elan import write_eaf

        segments = SegmentList(
            [
                Segment(0, 1, "CONTEXT1"),
                Segment(1, 2, "TARGET"),
                Segment(2, 3, "CONTEXT2"),
                Segment(5, 6, "TARGET"),  # Second occurrence
            ]
        )

        eaf = tmp_path / "test.eaf"
        write_eaf(segments, eaf)

        examples = find_gloss_examples(
            gloss="TARGET",
            elan_files=[eaf],
            max_examples=10,
        )

        assert len(examples) == 2
        assert all(ex["gloss"].upper() == "TARGET" for ex in examples)

    def test_case_insensitive(self, tmp_path):
        """Test case-insensitive search."""
        from sltk.io.elan import write_eaf

        segments = SegmentList(
            [
                Segment(0, 1, "Hello"),
            ]
        )

        eaf = tmp_path / "test.eaf"
        write_eaf(segments, eaf)

        examples = find_gloss_examples(
            gloss="hello",  # lowercase
            elan_files=[eaf],
        )

        assert len(examples) == 1


class TestComputePoseCoverage:
    """Tests for pose coverage computation."""

    def test_full_coverage(self, tmp_path):
        """Test with full coverage (all frames valid)."""
        from sltk.data.core import PoseSequence
        from sltk.io.h5 import save_poses_h5

        poses = PoseSequence(
            data=np.ones((100, 21, 3), dtype=np.float32),  # All non-zero
            fps=25.0,
            format="test",
        )

        h5_path = tmp_path / "poses.h5"
        save_poses_h5(poses, h5_path)

        coverage = compute_pose_coverage(h5_path)

        assert coverage["total_frames"] == 100
        assert coverage["valid_frames"] == 100
        assert coverage["coverage_ratio"] == 1.0
        assert coverage["num_gaps"] == 0

    def test_partial_coverage(self, tmp_path):
        """Test with partial coverage (some zero frames)."""
        from sltk.data.core import PoseSequence
        from sltk.io.h5 import save_poses_h5

        data = np.ones((100, 21, 3), dtype=np.float32)
        data[20:30] = 0  # Gap of 10 frames

        poses = PoseSequence(
            data=data,
            fps=25.0,
            format="test",
        )

        h5_path = tmp_path / "poses.h5"
        save_poses_h5(poses, h5_path)

        coverage = compute_pose_coverage(h5_path)

        assert coverage["total_frames"] == 100
        assert coverage["valid_frames"] == 90
        assert coverage["coverage_ratio"] == 0.9
        assert coverage["num_gaps"] >= 1


class TestComputeDatasetStatistics:
    """Tests for compute_dataset_statistics function."""

    def test_basic_statistics(self, tmp_path):
        """Test basic statistics computation from ELAN files."""
        from sltk.io.elan import write_eaf
        from sltk.utils.research import compute_dataset_statistics

        # Create test ELAN files
        segments1 = SegmentList(
            [
                Segment(0, 1, "HELLO", tier="Gloss"),
                Segment(1, 2, "WORLD", tier="Gloss"),
            ]
        )
        segments2 = SegmentList(
            [
                Segment(0, 0.5, "HELLO", tier="Gloss"),
                Segment(1, 1.5, "FOO", tier="Gloss"),
            ]
        )

        eaf1 = tmp_path / "test1.eaf"
        eaf2 = tmp_path / "test2.eaf"
        write_eaf(segments1, eaf1)
        write_eaf(segments2, eaf2)

        stats = compute_dataset_statistics([eaf1, eaf2], name="test")

        assert stats.name == "test"
        assert stats.num_samples == 2
        assert stats.num_segments == 4
        assert stats.vocabulary_size == 3  # HELLO, WORLD, FOO
        assert "HELLO" in stats.gloss_frequencies

    def test_empty_files_list(self):
        """Test with empty files list."""
        from sltk.utils.research import compute_dataset_statistics

        stats = compute_dataset_statistics([], name="empty")

        assert stats.name == "empty"
        assert stats.num_samples == 0
        assert stats.num_segments == 0

    def test_tier_counts(self, tmp_path):
        """Test tier counting."""
        from sltk.io.elan import write_eaf
        from sltk.utils.research import compute_dataset_statistics

        segments = SegmentList(
            [
                Segment(0, 1, "A", tier="Tier1"),
                Segment(1, 2, "B", tier="Tier1"),
                Segment(2, 3, "C", tier="Tier2"),
            ]
        )

        eaf = tmp_path / "test.eaf"
        write_eaf(segments, eaf)

        stats = compute_dataset_statistics([eaf], name="test")

        assert "Tier1" in stats.tier_counts
        assert "Tier2" in stats.tier_counts
        assert stats.tier_counts["Tier1"] == 2
        assert stats.tier_counts["Tier2"] == 1

    def test_gloss_frequency_limit(self, tmp_path):
        """Test top_k_glosses parameter."""
        from sltk.io.elan import write_eaf
        from sltk.utils.research import compute_dataset_statistics

        # Create many different glosses
        segments = SegmentList(
            [Segment(i, i + 0.5, f"GLOSS_{i % 100}", tier="Gloss") for i in range(200)]
        )

        eaf = tmp_path / "test.eaf"
        write_eaf(segments, eaf)

        stats = compute_dataset_statistics([eaf], name="test", top_k_glosses=10)

        # Should only have top 10 glosses
        assert len(stats.gloss_frequencies) <= 10


class TestBatchExtractPoses:
    """Tests for batch_extract_poses function."""

    def test_skip_existing(self, tmp_path):
        """Test that existing files are skipped."""
        from sltk.utils.research import batch_extract_poses

        # Create some video files
        videos = [tmp_path / f"video_{i}.mp4" for i in range(3)]
        for v in videos:
            v.touch()

        # Create existing output file
        output_dir = tmp_path / "output"
        output_dir.mkdir()
        existing = output_dir / "video_0_mediapipe.h5"
        existing.touch()

        # batch_extract_poses requires actual extractor setup
        # We can at least test the interface
        try:
            result = batch_extract_poses(
                video_paths=videos,
                output_dir=output_dir,
                extractor="mediapipe",
                skip_existing=True,
            )
            # If successful, check skipped count
            assert "skipped" in result
        except (ImportError, RuntimeError):
            # Expected if mediapipe not installed
            pass

    def test_unknown_extractor_raises(self, tmp_path):
        """Test that unknown extractor raises ConfigurationError."""
        from sltk.utils.research import batch_extract_poses

        videos = [tmp_path / "video.mp4"]
        videos[0].touch()

        with pytest.raises(ConfigurationError, match="Unknown extractor"):
            batch_extract_poses(
                video_paths=videos,
                output_dir=tmp_path / "output",
                extractor="nonexistent_extractor",
            )


class TestAlignPosesToSegments:
    """Tests for align_poses_to_segments function."""

    def test_basic_alignment(self):
        """Test basic pose-segment alignment."""
        from sltk.data.core import PoseSequence
        from sltk.utils.research import align_poses_to_segments

        data = np.random.randn(100, 21, 3).astype(np.float32)
        poses = PoseSequence(data=data, fps=25.0, format="test")

        segments = SegmentList(
            [
                Segment(0.0, 1.0, "A"),  # frames 0-25
                Segment(2.0, 3.0, "B"),  # frames 50-75
            ]
        )

        clips = list(align_poses_to_segments(poses, segments))

        assert len(clips) == 2
        seg1, clip1 = clips[0]
        seg2, clip2 = clips[1]

        assert seg1.label == "A"
        assert clip1.shape[0] == 25  # 1 second at 25fps

        assert seg2.label == "B"
        assert clip2.shape[0] == 25

    def test_with_padding(self):
        """Test alignment with padding frames."""
        from sltk.data.core import PoseSequence
        from sltk.utils.research import align_poses_to_segments

        data = np.random.randn(100, 21, 3).astype(np.float32)
        poses = PoseSequence(data=data, fps=25.0, format="test")

        segments = SegmentList(
            [
                Segment(1.0, 2.0, "A"),  # frames 25-50
            ]
        )

        clips = list(align_poses_to_segments(poses, segments, padding_frames=5))

        assert len(clips) == 1
        seg, clip = clips[0]

        # Should have 25 frames + 5 padding on each side = 35 frames
        assert clip.shape[0] == 35

    def test_segment_at_boundary(self):
        """Test alignment when segment is at start boundary."""
        from sltk.data.core import PoseSequence
        from sltk.utils.research import align_poses_to_segments

        data = np.random.randn(50, 21, 3).astype(np.float32)
        poses = PoseSequence(data=data, fps=25.0, format="test")

        segments = SegmentList(
            [
                Segment(0.0, 0.5, "START"),  # frames 0-12
            ]
        )

        clips = list(align_poses_to_segments(poses, segments, padding_frames=5))

        assert len(clips) == 1
        seg, clip = clips[0]

        # Padding should be clipped to available frames
        assert clip.shape[0] <= 17  # 12 + 5 padding

    def test_empty_segments(self):
        """Test alignment with empty segment list."""
        from sltk.data.core import PoseSequence
        from sltk.utils.research import align_poses_to_segments

        data = np.random.randn(100, 21, 3).astype(np.float32)
        poses = PoseSequence(data=data, fps=25.0, format="test")

        segments = SegmentList([])

        clips = list(align_poses_to_segments(poses, segments))

        assert len(clips) == 0


class TestExportFormats:
    """Tests for different export formats."""

    def test_jsonl_export(self, tmp_path):
        """Test JSONL export format."""
        from sltk.io.elan import write_eaf

        segments = SegmentList(
            [
                Segment(0, 1, "A"),
                Segment(1, 2, "B"),
            ]
        )

        eaf = tmp_path / "test.eaf"
        write_eaf(segments, eaf)

        output = tmp_path / "output.jsonl"
        export_segments_for_training(
            elan_files=[eaf],
            output_path=output,
            format="jsonl",
        )

        assert output.exists()

        # Check JSONL format (one JSON object per line)
        with open(output) as f:
            lines = f.readlines()

        assert len(lines) == 2
        for line in lines:
            data = json.loads(line)
            assert "label" in data

    def test_invalid_format_raises(self, tmp_path):
        """Test that invalid format raises error."""
        from sltk.io.elan import write_eaf

        segments = SegmentList([Segment(0, 1, "A")])
        eaf = tmp_path / "test.eaf"
        write_eaf(segments, eaf)

        from sltk.exceptions import UnsupportedFormatError

        with pytest.raises(UnsupportedFormatError, match="invalid_format"):
            export_segments_for_training(
                elan_files=[eaf],
                output_path=tmp_path / "output.xyz",
                format="invalid_format",
            )


class TestDatasetStatsEdgeCases:
    """Edge case tests for DatasetStats."""

    def test_empty_gloss_frequencies(self):
        """Test DatasetStats with no gloss frequencies."""
        stats = DatasetStats(name="test")

        assert stats.gloss_frequencies == {}

    def test_all_fields(self):
        """Test DatasetStats with all fields populated."""
        stats = DatasetStats(
            name="complete",
            num_samples=100,
            num_segments=500,
            total_duration_sec=3600.0,
            vocabulary_size=200,
            avg_segment_duration=1.5,
            segment_duration_std=0.5,
            glosses_per_sample_avg=5.0,
            tier_counts={"Gloss": 400, "Translation": 100},
            gloss_frequencies={"HELLO": 50, "WORLD": 30},
            duration_histogram=[10, 20, 30],
            signers=["signer1", "signer2"],
            num_signers=2,
        )

        d = stats.to_dict()
        assert d["name"] == "complete"
        assert d["num_samples"] == 100
        assert len(d["signers"]) == 2


class TestVocabularyEdgeCases:
    """Edge case tests for vocabulary functions."""

    def test_empty_vocabulary_comparison(self):
        """Test comparing empty vocabularies."""
        stats1 = DatasetStats(name="A", gloss_frequencies={})
        stats2 = DatasetStats(name="B", gloss_frequencies={})

        result = compare_vocabularies(stats1, stats2)

        assert result["intersection_size"] == 0
        assert result["jaccard_index"] == 0  # 0/0 case

    def test_one_empty_vocabulary(self):
        """Test comparing when one vocabulary is empty."""
        stats1 = DatasetStats(name="A", gloss_frequencies={"HELLO": 1})
        stats2 = DatasetStats(name="B", gloss_frequencies={})

        result = compare_vocabularies(stats1, stats2)

        assert result["intersection_size"] == 0
        assert result["overlap_ratio_2"] == 0  # Empty vocab



"""
GPU-Accelerated ViTDet Batch Processor for WiLoR Hand Pose Estimation

Optimized for NVIDIA H200 GPU with 141GB HBM3 memory.

Key Optimizations:
    1. Full GPU pipeline - no CPU-GPU transfers during processing
    2. Batched grid_sample for efficient affine crops
    3. ROI-based Gaussian blur (not full image)
    4. Vectorized operations throughout
    5. Pre-allocated output buffers
    6. Memory-efficient chunked processing for large batches

Author: Optimized for sign language video processing pipeline
"""

from dataclasses import dataclass

import numpy as np
import torch
import torch.nn.functional as F


@dataclass
class ViTDetConfig:
    """Configuration for ViTDet batch processor."""

    img_size: int = 256
    bbox_shape: tuple[int, int] | None = None
    image_mean: tuple[float, float, float] = (0.485, 0.456, 0.406)
    image_std: tuple[float, float, float] = (0.229, 0.224, 0.225)
    rescale_factor: float = 2.5

    @classmethod
    def from_cfg(cls, cfg) -> "ViTDetConfig":
        """Create config from yacs CfgNode."""
        return cls(
            img_size=cfg.MODEL.IMAGE_SIZE,
            bbox_shape=cfg.MODEL.get("BBOX_SHAPE", None),
            image_mean=tuple(cfg.MODEL.IMAGE_MEAN),
            image_std=tuple(cfg.MODEL.IMAGE_STD),
        )


class ViTDetGPUProcessor:
    """
    GPU-accelerated batch processor for hand detection crops.

    Replaces ViTDetDataset/ViTDetDatasetBatch with a fully GPU-based pipeline.
    Designed for maximum throughput on H200 GPUs.

    Features:
        - Batched affine transforms via grid_sample
        - Efficient Gaussian blur using separable convolution
        - Vectorized normalization
        - Zero CPU-GPU transfers during processing
        - Support for variable detection counts per frame

    Example:
        >>> processor = ViTDetGPUProcessor(config, device)
        >>> batch = processor(frames, boxes, is_right, img_indices, confidences)
        >>> # batch['img'] is ready for WiLoR model input
    """

    def __init__(
        self,
        config: ViTDetConfig,
        device: torch.device,
        max_kernel_size: int = 31,
    ):
        """
        Initialize the GPU processor.

        Args:
            config: ViTDetConfig with model parameters
            device: Target CUDA device
            max_kernel_size: Maximum Gaussian kernel size to pre-compute
        """
        self.config = config
        self.device = device
        self.img_size = config.img_size
        self.bbox_shape = config.bbox_shape

        # Pre-compute normalization tensors on GPU
        # Shape: (1, 3, 1, 1) for broadcasting over (N, C, H, W)
        self.mean = (
            torch.tensor(config.image_mean, device=device, dtype=torch.float32).view(1, 3, 1, 1)
            * 255.0
        )

        self.std = (
            torch.tensor(config.image_std, device=device, dtype=torch.float32).view(1, 3, 1, 1)
            * 255.0
        )

        # Pre-compute Gaussian kernels for common sigma values
        # This avoids kernel computation during inference
        self._precompute_gaussian_kernels(max_kernel_size)

    def _precompute_gaussian_kernels(self, max_size: int):
        """
        Pre-compute 1D Gaussian kernels for separable convolution.

        Separable 2D Gaussian = 1D horizontal * 1D vertical
        This is O(2n) vs O(n²) for direct 2D convolution.
        """
        self.gaussian_kernels = {}

        for ksize in range(3, max_size + 1, 2):  # Odd kernel sizes only
            sigma = (ksize - 1) / 6.0  # Approximate sigma from kernel size

            # 1D Gaussian kernel
            x = torch.arange(ksize, device=self.device, dtype=torch.float32)
            x = x - (ksize - 1) / 2
            kernel_1d = torch.exp(-(x**2) / (2 * sigma**2))
            kernel_1d = kernel_1d / kernel_1d.sum()

            # Store as (1, 1, 1, ksize) for F.conv2d
            self.gaussian_kernels[ksize] = kernel_1d.view(1, 1, 1, ksize)

    def _get_gaussian_kernel(self, sigma: float) -> tuple[torch.Tensor, int]:
        """
        Get pre-computed Gaussian kernel for given sigma.

        Args:
            sigma: Standard deviation of Gaussian

        Returns:
            Tuple of (1D kernel tensor, kernel size)
        """
        # Compute kernel size from sigma (covers 99.7% of distribution)
        ksize = int(np.ceil(sigma * 6)) | 1  # Ensure odd
        ksize = max(3, min(ksize, max(self.gaussian_kernels.keys())))

        if ksize not in self.gaussian_kernels:
            # Fallback: compute on-the-fly (rare case)
            x = torch.arange(ksize, device=self.device, dtype=torch.float32)
            x = x - (ksize - 1) / 2
            kernel_1d = torch.exp(-(x**2) / (2 * sigma**2))
            kernel_1d = kernel_1d / kernel_1d.sum()
            return kernel_1d.view(1, 1, 1, ksize), ksize

        return self.gaussian_kernels[ksize], ksize

    @torch.no_grad()
    def _apply_gaussian_blur_batched(
        self,
        images: torch.Tensor,
        sigma: float,
    ) -> torch.Tensor:
        """
        Apply Gaussian blur using separable convolution.

        This is more efficient than 2D convolution:
        - 2D conv: O(N * C * H * W * K²)
        - Separable: O(N * C * H * W * 2K)

        Args:
            images: (N, C, H, W) tensor
            sigma: Gaussian standard deviation

        Returns:
            Blurred images (N, C, H, W)
        """
        if sigma < 0.55:
            return images

        kernel_1d, ksize = self._get_gaussian_kernel(sigma)
        pad = ksize // 2

        N, C, H, W = images.shape

        # Reshape for grouped convolution: (N*C, 1, H, W)
        # Use reshape instead of view to handle non-contiguous tensors
        x = images.reshape(N * C, 1, H, W)

        # Horizontal pass
        x = F.pad(x, (pad, pad, 0, 0), mode="reflect")
        x = F.conv2d(x, kernel_1d, groups=1)

        # Vertical pass (transpose kernel)
        x = F.pad(x, (0, 0, pad, pad), mode="reflect")
        x = F.conv2d(x, kernel_1d.transpose(2, 3), groups=1)

        return x.reshape(N, C, H, W)

    @torch.no_grad()
    def _compute_bbox_geometry(
        self,
        boxes: torch.Tensor,
        rescale_factor: float,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Compute centers and sizes from bounding boxes.

        Vectorized computation for all detections simultaneously.

        Args:
            boxes: (M, 4) tensor of [x1, y1, x2, y2]
            rescale_factor: Scale factor for bbox expansion

        Returns:
            centers: (M, 2) tensor of [cx, cy]
            bbox_sizes: (M,) tensor of bbox sizes
        """
        # Compute centers: midpoint of bbox
        centers = (boxes[:, 2:4] + boxes[:, 0:2]) / 2.0

        # Compute scales (normalized by 200 as per original WiLoR convention)
        scales = rescale_factor * (boxes[:, 2:4] - boxes[:, 0:2]) / 200.0

        # Apply aspect ratio expansion if configured
        if self.bbox_shape is not None:
            target_ratio = self.bbox_shape[0] / self.bbox_shape[1]
            scales_200 = scales * 200

            current_ratio = scales_200[:, 0] / (scales_200[:, 1] + 1e-8)

            # Expand width or height to match target aspect ratio
            needs_width_expand = current_ratio < target_ratio

            # Clone to avoid in-place modification issues
            scales_200 = scales_200.clone()
            scales_200[needs_width_expand, 0] = scales_200[needs_width_expand, 1] * target_ratio
            scales_200[~needs_width_expand, 1] = scales_200[~needs_width_expand, 0] / target_ratio

            bbox_sizes = scales_200.max(dim=1).values
        else:
            bbox_sizes = (scales * 200).max(dim=1).values

        return centers, bbox_sizes

    @torch.no_grad()
    def _build_affine_grids(
        self,
        centers: torch.Tensor,
        bbox_sizes: torch.Tensor,
        is_right: torch.Tensor,
        img_h: int,
        img_w: int,
    ) -> torch.Tensor:
        """
        Build affine sampling grids for all detections.

        Uses F.affine_grid to create sampling coordinates that:
        1. Center the crop on the detection
        2. Scale to the target patch size
        3. Flip horizontally for left hands

        Args:
            centers: (M, 2) detection centers [cx, cy]
            bbox_sizes: (M,) bounding box sizes
            is_right: (M,) 1.0 for right hand, 0.0 for left
            img_h, img_w: Source image dimensions

        Returns:
            Sampling grid (M, img_size, img_size, 2) in [-1, 1] coordinates
        """
        M = centers.shape[0]

        # Build 2x3 affine matrices
        # grid_sample expects: x_src = theta[0,0]*x_dst + theta[0,1]*y_dst + theta[0,2]
        #                      y_src = theta[1,0]*x_dst + theta[1,1]*y_dst + theta[1,2]

        theta = torch.zeros(M, 2, 3, device=self.device, dtype=torch.float32)

        # Scale factors: map [-1,1] grid to bbox region in normalized image coords
        # bbox_size in pixels -> fraction of image
        sx = bbox_sizes / img_w
        sy = bbox_sizes / img_h

        # Translation: center of bbox in normalized [-1, 1] coordinates
        tx = (centers[:, 0] / img_w) * 2 - 1
        ty = (centers[:, 1] / img_h) * 2 - 1

        # Flip sign for left hands (is_right=0 -> flip_sign=-1)
        # Only flip the scale to mirror content, NOT the translation.
        # The crop should still be centered on the bbox, just with mirrored content.
        flip_sign = 2 * is_right - 1

        # Affine matrix components
        theta[:, 0, 0] = sx * flip_sign  # Scale X with flip (mirrors content for left hands)
        theta[:, 1, 1] = sy  # Scale Y
        theta[:, 0, 2] = tx  # Translate X (NOT flipped - crop stays at bbox center)
        theta[:, 1, 2] = ty  # Translate Y

        # Generate sampling grid
        grid = F.affine_grid(
            theta,
            size=(M, 1, self.img_size, self.img_size),
            align_corners=False,
        )

        return grid

    @torch.no_grad()
    def __call__(
        self,
        images: torch.Tensor,
        boxes: torch.Tensor,
        is_right: torch.Tensor,
        img_indices: torch.Tensor,
        confidences: torch.Tensor,
        rescale_factor: float | None = None,
    ) -> dict[str, torch.Tensor]:
        """
        Extract and preprocess hand crops for WiLoR model.

        Full GPU pipeline with no CPU transfers.

        Args:
            images: (N, H, W, 3) uint8 tensor - batch of video frames on GPU
            boxes: (M, 4) float32 tensor - bounding boxes [x1, y1, x2, y2]
            is_right: (M,) float32 tensor - 1.0 for right hand, 0.0 for left
            img_indices: (M,) int32/int64 tensor - source frame index for each detection
            confidences: (M,) float32 tensor - YOLO detection confidence scores
            rescale_factor: Optional override for bbox expansion factor

        Returns:
            Dictionary with preprocessed batch ready for WiLoR:
                - img: (M, 3, img_size, img_size) normalized patches
                - box_center: (M, 2) detection centers
                - box_size: (M,) bbox sizes
                - right: (M,) handedness flags
                - img_idx: (M,) source frame indices
                - img_size: (M, 2) source image dimensions [W, H]
                - confidence: (M,) detection confidences
        """
        if rescale_factor is None:
            rescale_factor = self.config.rescale_factor

        M = boxes.shape[0]

        # Handle empty detection case
        if M == 0:
            return self._create_empty_batch(images.shape[1], images.shape[2])

        # Ensure all inputs are on correct device
        boxes = boxes.to(self.device, dtype=torch.float32)
        is_right = is_right.to(self.device, dtype=torch.float32)
        img_indices = img_indices.to(self.device, dtype=torch.int64)
        confidences = confidences.to(self.device, dtype=torch.float32)

        # Get image dimensions
        N, H, W, C = images.shape

        # Convert images: (N, H, W, 3) -> (N, 3, H, W) float32
        # This is the main format conversion - done once
        images_nchw = images.permute(0, 3, 1, 2).to(dtype=torch.float32)

        # Compute bbox geometry (vectorized)
        centers, bbox_sizes = self._compute_bbox_geometry(boxes, rescale_factor)

        # Compute antialiasing blur sigma
        # Based on downsampling factor (original image -> patch size)
        downsample_factors = bbox_sizes / self.img_size
        max_downsample = downsample_factors.max().item()
        sigma = max(0, (max_downsample - 1) / 2)

        # Apply Gaussian blur for antialiasing if needed
        if sigma > 0.55:
            images_nchw = self._apply_gaussian_blur_batched(images_nchw, sigma)

        # Gather source images for each detection
        # This expands images from (N, C, H, W) to (M, C, H, W)
        src_images = images_nchw[img_indices]

        # Build affine sampling grids for all detections
        grids = self._build_affine_grids(centers, bbox_sizes, is_right, H, W)

        # Extract patches via grid_sample (batched bilinear interpolation)
        patches = F.grid_sample(
            src_images,
            grids,
            mode="bilinear",
            padding_mode="zeros",
            align_corners=False,
        )

        # Apply normalization: (patch - mean) / std
        patches = (patches - self.mean) / self.std

        # Build output dictionary matching original ViTDetDataset format
        return {
            "img": patches,  # (M, 3, H, W)
            "box_center": centers,  # (M, 2)
            "box_size": bbox_sizes,  # (M,)
            "right": is_right,  # (M,)
            "img_idx": img_indices,  # (M,)
            "img_size": torch.tensor(  # (M, 2)
                [[W, H]], device=self.device, dtype=torch.float32
            ).expand(M, 2),
            "confidence": confidences,  # (M,)
        }

    def _create_empty_batch(self, img_h: int, img_w: int) -> dict[str, torch.Tensor]:
        """Create empty batch structure for zero-detection case."""
        return {
            "img": torch.empty(
                (0, 3, self.img_size, self.img_size), device=self.device, dtype=torch.float32
            ),
            "box_center": torch.empty((0, 2), device=self.device, dtype=torch.float32),
            "box_size": torch.empty((0,), device=self.device, dtype=torch.float32),
            "right": torch.empty((0,), device=self.device, dtype=torch.float32),
            "img_idx": torch.empty((0,), device=self.device, dtype=torch.int64),
            "img_size": torch.empty((0, 2), device=self.device, dtype=torch.float32),
            "confidence": torch.empty((0,), device=self.device, dtype=torch.float32),
        }


class ViTDetGPUProcessorChunked(ViTDetGPUProcessor):
    """
    Memory-efficient chunked processor for very large detection batches.

    Processes detections in chunks to avoid OOM on extremely dense frames.
    Inherits all optimizations from ViTDetGPUProcessor.

    Use this when processing frames with potentially hundreds of detections.
    """

    def __init__(
        self,
        config: ViTDetConfig,
        device: torch.device,
        chunk_size: int = 512,
        **kwargs,
    ):
        """
        Initialize chunked processor.

        Args:
            config: ViTDetConfig with model parameters
            device: Target CUDA device
            chunk_size: Maximum detections to process per chunk
        """
        super().__init__(config, device, **kwargs)
        self.chunk_size = chunk_size

    @torch.no_grad()
    def __call__(
        self,
        images: torch.Tensor,
        boxes: torch.Tensor,
        is_right: torch.Tensor,
        img_indices: torch.Tensor,
        confidences: torch.Tensor,
        rescale_factor: float | None = None,
    ) -> dict[str, torch.Tensor]:
        """
        Process detections in memory-efficient chunks.

        Same interface as parent class, but processes in chunks internally.
        """
        M = boxes.shape[0]

        # For small batches, use parent implementation directly
        if M <= self.chunk_size:
            return super().__call__(
                images, boxes, is_right, img_indices, confidences, rescale_factor
            )

        # Process in chunks and concatenate
        all_outputs = []

        for start in range(0, M, self.chunk_size):
            end = min(start + self.chunk_size, M)

            chunk_output = super().__call__(
                images,
                boxes[start:end],
                is_right[start:end],
                img_indices[start:end],
                confidences[start:end],
                rescale_factor,
            )
            all_outputs.append(chunk_output)

        # Concatenate all chunks
        return {
            key: torch.cat([out[key] for out in all_outputs], dim=0)
            for key in all_outputs[0].keys()
        }



"""Tests for feature discovery utilities."""

from __future__ import annotations

from pathlib import Path

import pytest

from sltk.data.feature_discovery import (
    detect_naming_pattern,
    discover_features,
    extract_sample_id,
    list_sample_ids,
    scan_feature_directory,
)


class TestDiscoverFeatures:
    """Tests for discover_features function."""

    @pytest.fixture
    def feature_dir(self, tmp_path):
        """Create temp directory with feature files."""
        feat_dir = tmp_path / "features"
        feat_dir.mkdir()

        # Create ID-based files
        (feat_dir / "sample_001_wilor.h5").touch()
        (feat_dir / "sample_001_smpl.h5").touch()
        (feat_dir / "sample_002_wilor.h5").touch()

        return feat_dir

    def test_discover_existing_features(self, feature_dir):
        """Test discovering existing features."""
        found = discover_features(
            sample_id="sample_001",
            source_path=feature_dir,
            feature_types=["wilor", "smpl"],
            naming_pattern="{sample_id}_{feature_type}.h5",
        )

        assert "wilor" in found
        assert "smpl" in found
        assert found["wilor"].endswith("sample_001_wilor.h5")

    def test_discover_missing_features(self, feature_dir):
        """Test discovering when feature is missing."""
        found = discover_features(
            sample_id="sample_001",
            source_path=feature_dir,
            feature_types=["mediapipe"],  # Not in directory
            naming_pattern="{sample_id}_{feature_type}.h5",
        )

        assert "mediapipe" not in found

    def test_discover_nonexistent_sample(self, feature_dir):
        """Test discovering features for nonexistent sample."""
        found = discover_features(
            sample_id="nonexistent",
            source_path=feature_dir,
            feature_types=["wilor"],
            naming_pattern="{sample_id}_{feature_type}.h5",
        )

        assert len(found) == 0

    def test_discover_nonexistent_dir(self, tmp_path):
        """Test discovering in nonexistent directory."""
        found = discover_features(
            sample_id="sample_001",
            source_path=tmp_path / "nonexistent",
            feature_types=["wilor"],
            naming_pattern="{sample_id}_{feature_type}.h5",
        )

        assert len(found) == 0


class TestDetectNamingPattern:
    """Tests for detect_naming_pattern function."""

    @pytest.fixture
    def id_based_dir(self, tmp_path):
        """Create directory with ID-based naming."""
        feat_dir = tmp_path / "features"
        feat_dir.mkdir()

        for i in range(10):
            (feat_dir / f"sample_{i:03d}_wilor.h5").touch()

        return feat_dir

    @pytest.fixture
    def gloss_based_dir(self, tmp_path):
        """Create directory with gloss-based naming."""
        feat_dir = tmp_path / "features"
        feat_dir.mkdir()

        for gloss in ["HELLO", "WORLD", "SIGN"]:
            (feat_dir / f"asldict-{gloss}_wilor.h5").touch()

        return feat_dir

    def test_detect_id_based_pattern(self, id_based_dir):
        """Test detecting ID-based naming pattern."""
        patterns = detect_naming_pattern(id_based_dir)

        assert len(patterns) > 0
        # Should detect the {sample_id}_{feature_type}.{ext} pattern
        pattern_strs = [p["pattern"] for p in patterns]
        assert any("sample_id" in p for p in pattern_strs)

    def test_detect_gloss_based_pattern(self, gloss_based_dir):
        """Test detecting gloss-based naming pattern."""
        patterns = detect_naming_pattern(gloss_based_dir)

        assert len(patterns) > 0

    def test_detect_empty_dir(self, tmp_path):
        """Test detecting patterns in empty directory."""
        empty_dir = tmp_path / "empty"
        empty_dir.mkdir()

        patterns = detect_naming_pattern(empty_dir)
        assert len(patterns) == 0

    def test_detect_nonexistent_dir(self, tmp_path):
        """Test detecting patterns in nonexistent directory."""
        patterns = detect_naming_pattern(tmp_path / "nonexistent")
        assert len(patterns) == 0


class TestExtractSampleId:
    """Tests for extract_sample_id function."""

    def test_extract_id_based(self):
        """Test extracting from ID-based naming."""
        sample_id = extract_sample_id(
            "sample_001_wilor.h5",
            "{sample_id}_{feature_type}.h5",
        )
        assert sample_id == "sample_001"

    def test_extract_gloss_based(self):
        """Test extracting from gloss-based naming."""
        sample_id = extract_sample_id(
            "asldict-HELLO_wilor.h5",
            "{source}-{gloss}_{feature_type}.h5",
        )
        # Should construct sample_id from source-gloss
        assert sample_id == "asldict-HELLO"

    def test_extract_no_match(self):
        """Test extracting from non-matching filename."""
        sample_id = extract_sample_id(
            "random_file.txt",
            "{sample_id}_{feature_type}.h5",
        )
        assert sample_id is None

    def test_extract_flat_pattern(self):
        """Test extracting from flat naming pattern."""
        sample_id = extract_sample_id(
            "sample_001.h5",
            "{sample_id}.h5",
        )
        assert sample_id == "sample_001"


class TestListSampleIds:
    """Tests for list_sample_ids function."""

    @pytest.fixture
    def feature_dir(self, tmp_path):
        """Create directory with multiple samples."""
        feat_dir = tmp_path / "features"
        feat_dir.mkdir()

        samples = ["sample_001", "sample_002", "sample_003"]
        for sample in samples:
            (feat_dir / f"{sample}_wilor.h5").touch()
            (feat_dir / f"{sample}_smpl.h5").touch()

        return feat_dir

    def test_list_unique_sample_ids(self, feature_dir):
        """Test listing unique sample IDs."""
        sample_ids = list_sample_ids(
            feature_dir,
            "{sample_id}_{feature_type}.h5",
        )

        assert len(sample_ids) == 3
        assert "sample_001" in sample_ids
        assert "sample_002" in sample_ids
        assert "sample_003" in sample_ids

    def test_list_empty_dir(self, tmp_path):
        """Test listing in empty directory."""
        empty_dir = tmp_path / "empty"
        empty_dir.mkdir()

        sample_ids = list_sample_ids(
            empty_dir,
            "{sample_id}_{feature_type}.h5",
        )

        assert len(sample_ids) == 0

    def test_list_nonexistent_dir(self, tmp_path):
        """Test listing in nonexistent directory."""
        sample_ids = list_sample_ids(
            tmp_path / "nonexistent",
            "{sample_id}_{feature_type}.h5",
        )

        assert len(sample_ids) == 0


class TestScanFeatureDirectory:
    """Tests for scan_feature_directory function."""

    @pytest.fixture
    def mixed_feature_dir(self, tmp_path):
        """Create directory with multiple feature types."""
        feat_dir = tmp_path / "features"
        feat_dir.mkdir()

        # Create wilor files
        (feat_dir / "sample_001_wilor.h5").touch()
        (feat_dir / "sample_002_wilor.h5").touch()

        # Create smpl files
        (feat_dir / "sample_001_smpl.h5").touch()

        # Create mediapipe files
        (feat_dir / "sample_001_mediapipe.h5").touch()

        return feat_dir

    def test_scan_all_types(self, mixed_feature_dir):
        """Test scanning for all feature types."""
        result = scan_feature_directory(mixed_feature_dir)

        assert "wilor" in result
        assert len(result["wilor"]) == 2
        assert "smpl" in result
        assert len(result["smpl"]) == 1

    def test_scan_specific_type(self, mixed_feature_dir):
        """Test scanning for specific feature type."""
        result = scan_feature_directory(mixed_feature_dir, feature_type="wilor")

        assert "wilor" in result
        assert len(result["wilor"]) == 2
        assert "smpl" not in result

    def test_scan_empty_dir(self, tmp_path):
        """Test scanning empty directory."""
        empty_dir = tmp_path / "empty"
        empty_dir.mkdir()

        result = scan_feature_directory(empty_dir)
        assert len(result) == 0

    def test_scan_nonexistent_dir(self, tmp_path):
        """Test scanning nonexistent directory."""
        result = scan_feature_directory(tmp_path / "nonexistent")
        assert len(result) == 0


class TestFuzzyMatching:
    """Tests for fuzzy feature file matching."""

    @pytest.fixture
    def varied_names_dir(self, tmp_path):
        """Create directory with varied naming conventions."""
        feat_dir = tmp_path / "features"
        feat_dir.mkdir()

        # Different separators
        (feat_dir / "sample_001-wilor.h5").touch()
        (feat_dir / "sample_002.wilor.h5").touch()

        # Different extensions
        (feat_dir / "sample_003_wilor.pt").touch()
        (feat_dir / "sample_004_wilor.lzma").touch()

        return feat_dir

    def test_discover_with_different_separators(self, varied_names_dir):
        """Test discovering with different separators."""
        found = discover_features(
            sample_id="sample_001",
            source_path=varied_names_dir,
            feature_types=["wilor"],
            naming_pattern="{sample_id}_{feature_type}.h5",
        )

        # Should find even with dash separator via fuzzy matching
        assert "wilor" in found

    def test_discover_different_extensions(self, varied_names_dir):
        """Test discovering with different extensions."""
        found = discover_features(
            sample_id="sample_003",
            source_path=varied_names_dir,
            feature_types=["wilor"],
            naming_pattern="{sample_id}_{feature_type}.h5",
        )

        # Should find .pt file via fuzzy matching
        assert "wilor" in found



"""
Evaluation metrics for sign language tasks.
"""

from sltk.metrics.segmentation import (
    SegmentationResult,
    SegmentMatch,
    boundary_f1,
    boundary_f1_multi_tolerance,
    comprehensive_segmentation_metrics,
    confusion_matrix,
    frame_accuracy,
    label_accuracy,
    over_segmentation_rate,
    per_class_metrics,
    per_segment_iou,
    segment_iou,
    segmentation_metrics,
    segmentation_summary,
    under_segmentation_rate,
)
from sltk.metrics.translation import bleu_score, rouge_score, translation_metrics

__all__ = [
    # Translation metrics
    "translation_metrics",
    "bleu_score",
    "rouge_score",
    # Basic segmentation metrics
    "segmentation_metrics",
    "boundary_f1",
    "segment_iou",
    "frame_accuracy",
    # Comprehensive segmentation metrics
    "comprehensive_segmentation_metrics",
    "SegmentationResult",
    "SegmentMatch",
    "boundary_f1_multi_tolerance",
    "label_accuracy",
    "over_segmentation_rate",
    "under_segmentation_rate",
    "per_segment_iou",
    "per_class_metrics",
    "confusion_matrix",
    "segmentation_summary",
]



"""
Extraction endpoints for SLTK API.

Handles pose extraction pipeline jobs.
"""

from __future__ import annotations

import logging
import os
import threading
import uuid
from datetime import datetime, timedelta
from pathlib import Path

from fastapi import APIRouter, BackgroundTasks, HTTPException, Request

from sltk.api.dependencies import rate_limit, validate_path
from sltk.api.models import (
    ExtractionConfig,
    ExtractionJob,
    ExtractionRequest,
    StageStatus,
)

router = APIRouter(prefix="/api/extraction", tags=["extraction"])
logger = logging.getLogger(__name__)

# ============================================================================
# Job Storage
# ============================================================================

# In-memory job storage (use Redis/DB in production)
# Thread safety: Use _jobs_lock when accessing jobs or job_logs
jobs: dict[str, ExtractionJob] = {}
_jobs_lock = threading.Lock()

# In-memory log storage for extraction jobs
job_logs: dict[str, list[str]] = {}
_logs_lock = threading.Lock()

# Job cleanup configuration
JOB_TTL_HOURS = 24  # How long to keep completed jobs
MAX_JOBS = 1000  # Maximum number of jobs to keep
JOB_CLEANUP_INTERVAL_SECONDS = 3600  # Run cleanup every hour


def add_job_log(job_id: str, message: str, level: str = "info"):
    """Add a log message for a job. Thread-safe."""
    timestamp = datetime.now().strftime("%H:%M:%S")
    log_entry = f"[{timestamp}] [{level.upper()}] {message}"
    with _logs_lock:
        if job_id not in job_logs:
            job_logs[job_id] = []
        job_logs[job_id].append(log_entry)
        # Keep only last 1000 log entries
        if len(job_logs[job_id]) > 1000:
            job_logs[job_id] = job_logs[job_id][-1000:]


def cleanup_old_jobs() -> int:
    """
    Remove jobs older than TTL and enforce max job limit.

    This function is thread-safe and uses the _jobs_lock and _logs_lock
    to prevent race conditions when cleaning up jobs and logs.

    Returns:
        Number of jobs removed.
    """
    now = datetime.now()
    ttl_cutoff = now - timedelta(hours=JOB_TTL_HOURS)

    jobs_to_remove: list[str] = []

    # Find jobs to remove based on TTL
    with _jobs_lock:
        for job_id, job in jobs.items():
            # Remove completed/failed/cancelled jobs older than TTL
            if job.status in ("completed", "failed", "cancelled"):
                if job.completed_at:
                    try:
                        completed_time = datetime.fromisoformat(job.completed_at)
                        if completed_time < ttl_cutoff:
                            jobs_to_remove.append(job_id)
                    except ValueError:
                        # Invalid timestamp, mark for removal
                        jobs_to_remove.append(job_id)

        # Remove old jobs
        for job_id in jobs_to_remove:
            jobs.pop(job_id, None)

        # If still over limit, remove oldest completed jobs
        if len(jobs) > MAX_JOBS:
            completed_jobs = [
                (jid, j)
                for jid, j in jobs.items()
                if j.status in ("completed", "failed", "cancelled")
            ]
            # Sort by completed_at (oldest first), handling None
            completed_jobs.sort(key=lambda x: x[1].completed_at or "")

            excess = len(jobs) - MAX_JOBS
            for job_id, _ in completed_jobs[:excess]:
                jobs.pop(job_id, None)
                if job_id not in jobs_to_remove:
                    jobs_to_remove.append(job_id)

    # Clean up logs for removed jobs
    with _logs_lock:
        for job_id in jobs_to_remove:
            job_logs.pop(job_id, None)

    return len(jobs_to_remove)


def _get_job(job_id: str) -> ExtractionJob | None:
    """Get a job by ID with thread safety."""
    with _jobs_lock:
        return jobs.get(job_id)


def _update_job(job_id: str, **kwargs) -> None:
    """Update job fields with thread safety."""
    with _jobs_lock:
        if job_id in jobs:
            job = jobs[job_id]
            for key, value in kwargs.items():
                setattr(job, key, value)


class ExtractionCancelled(Exception):
    """Exception raised when extraction is cancelled."""

    pass


def check_job_cancelled(job_id: str) -> bool:
    """Check if a job has been cancelled. Returns True if cancelled. Thread-safe."""
    with _jobs_lock:
        if job_id in jobs and jobs[job_id].status == "cancelled":
            return True
        return False


def raise_if_cancelled(job_id: str):
    """Raise ExtractionCancelled if the job has been cancelled."""
    if check_job_cancelled(job_id):
        raise ExtractionCancelled(f"Job {job_id} was cancelled by user")


def cleanup_gpu_memory():
    """
    Force cleanup of GPU memory after job cancellation or completion.

    This helps free VRAM that may be held by extractors when jobs are
    cancelled mid-extraction.
    """
    import gc

    gc.collect()

    try:
        import torch

        if torch.cuda.is_available():
            torch.cuda.synchronize()
            torch.cuda.empty_cache()
    except ImportError:
        pass  # torch not installed


# ============================================================================
# Endpoints
# ============================================================================


@router.post("/start", response_model=ExtractionJob)
@rate_limit("5/minute")
async def start_extraction(
    request: Request, body: ExtractionRequest, background_tasks: BackgroundTasks
):
    """
    Start an extraction pipeline job.

    This runs in the background and can be monitored via /api/extraction/status/{job_id}.

    Rate limited to 5 requests per minute (extraction is very expensive).
    """
    # Security: Validate all paths are within allowed directories
    input_path = validate_path(body.input_path, must_exist=True)
    output_root = validate_path(body.output_root, must_exist=False, allow_creation=True)

    # Also validate any specific video paths if provided
    if body.video_paths:
        for video_path in body.video_paths:
            validate_path(video_path, must_exist=True)

    # Create output directory (already validated)
    output_root.mkdir(parents=True, exist_ok=True)

    # Create job
    job_id = str(uuid.uuid4())[:8]
    job = ExtractionJob(
        job_id=job_id,
        status="pending",
        created_at=datetime.now().isoformat(),
        input_path=str(input_path),
        output_root=str(output_root),
        config=body.config,
    )
    with _jobs_lock:
        jobs[job_id] = job

    # Start background extraction
    background_tasks.add_task(
        run_extraction_pipeline,
        job_id,
        input_path,
        output_root,
        body.config,
        body.recursive,
        body.video_paths,
    )

    return job


@router.get("/status/{job_id}", response_model=ExtractionJob)
async def get_extraction_status(job_id: str):
    """Get status of an extraction job."""
    with _jobs_lock:
        if job_id not in jobs:
            raise HTTPException(404, f"Job not found: {job_id}")
        return jobs[job_id]


@router.post("/cancel/{job_id}")
async def cancel_extraction(job_id: str):
    """Cancel an extraction job and free GPU resources."""
    with _jobs_lock:
        if job_id not in jobs:
            raise HTTPException(404, f"Job not found: {job_id}")

        job = jobs[job_id]
        if job.status == "running":
            job.status = "cancelled"
    add_job_log(job_id, "Cancellation requested - stopping at next checkpoint", "warning")

    # Schedule GPU cleanup to run after the extraction threads have had a chance to stop
    # This runs in a separate thread to not block the response
    def delayed_cleanup():
        import time

        time.sleep(2)  # Wait for extraction threads to notice cancellation and exit
        cleanup_gpu_memory()
        add_job_log(job_id, "GPU memory cleanup completed", "info")

    cleanup_thread = threading.Thread(target=delayed_cleanup, daemon=True)
    cleanup_thread.start()

    return {"status": "cancelled", "job_id": job_id}


@router.get("/jobs", response_model=list[ExtractionJob])
async def list_extraction_jobs():
    """List all extraction jobs."""
    with _jobs_lock:
        return list(jobs.values())


@router.get("/logs/{job_id}")
async def get_extraction_logs(job_id: str, offset: int = 0, limit: int = 100):
    """Get logs for an extraction job."""
    with _jobs_lock:
        if job_id not in jobs:
            raise HTTPException(404, f"Job not found: {job_id}")

    with _logs_lock:
        logs = job_logs.get(job_id, [])
        total = len(logs)
        result_logs = logs[offset : offset + limit]

    return {
        "job_id": job_id,
        "logs": result_logs,
        "total": total,
        "offset": offset,
        "has_more": offset + limit < total,
    }


# ============================================================================
# Pipeline Implementation
# ============================================================================


async def run_extraction_pipeline(
    job_id: str,
    input_path: Path,
    output_root: Path,
    config: ExtractionConfig,
    recursive: bool,
    video_paths: list[str] | None = None,
):
    """
    Run the full extraction pipeline.

    This is executed as a background task.
    """
    job = _get_job(job_id)
    if job is None:
        return
    _update_job(job_id, status="running", started_at=datetime.now().isoformat())

    add_job_log(job_id, "Starting extraction pipeline", "info")
    add_job_log(job_id, f"Input path: {input_path}", "info")
    add_job_log(job_id, f"Output root: {output_root}", "info")
    config_msg = (
        f"Config: mediapipe={config.enable_mediapipe}, "
        f"wilor={config.enable_wilor}, nlf={config.enable_nlf}"
    )
    add_job_log(job_id, config_msg, "info")

    try:
        # Use specific video paths if provided, otherwise discover
        if video_paths:
            video_files = [Path(p) for p in video_paths]
            add_job_log(job_id, f"Using {len(video_files)} specified video paths", "info")
        elif input_path.is_file():
            video_files = [input_path]
            add_job_log(job_id, "Processing single video file", "info")
        else:
            extensions = {".mp4", ".mov", ".avi", ".mkv"}
            video_files = []

            add_job_log(job_id, "Scanning directory for videos...", "info")

            def handle_error(err: OSError):
                add_job_log(job_id, f"Permission denied: {err.filename}", "warning")

            if recursive:
                for dirpath, dirnames, filenames in os.walk(input_path, onerror=handle_error):
                    dir_path = Path(dirpath)
                    for filename in filenames:
                        if any(filename.lower().endswith(ext) for ext in extensions):
                            video_files.append(dir_path / filename)
            else:
                try:
                    for item in input_path.iterdir():
                        if item.is_file() and any(
                            item.name.lower().endswith(ext) for ext in extensions
                        ):
                            video_files.append(item)
                except PermissionError:
                    add_job_log(job_id, f"Permission denied accessing {input_path}", "error")

        _update_job(job_id, total_videos=len(video_files))
        add_job_log(job_id, f"Found {len(video_files)} videos to process", "info")

        # Create output subdirectories
        poses_dir = output_root / "poses"
        features_dir = output_root / "features"
        segments_dir = output_root / "segments"

        poses_dir.mkdir(exist_ok=True)
        features_dir.mkdir(exist_ok=True)
        segments_dir.mkdir(exist_ok=True)

        add_job_log(job_id, "Output directories created", "info")

        # Process each video
        for i, video_path in enumerate(video_files):
            # Check cancellation status thread-safely
            if check_job_cancelled(job_id):
                add_job_log(job_id, "Job cancelled by user", "warning")
                break

            _update_job(job_id, current_video=video_path.name, processed_videos=i)

            add_job_log(job_id, f"Processing [{i+1}/{len(video_files)}]: {video_path.name}", "info")

            try:
                # Get fresh job reference for process_single_video
                current_job = _get_job(job_id)
                if current_job is None:
                    break
                await process_single_video(
                    video_path,
                    output_root,
                    config,
                    current_job,
                    job_id,
                )
                with _jobs_lock:
                    jobs[job_id].successful.append(str(video_path))
                add_job_log(job_id, f"Completed: {video_path.name}", "info")
            except ExtractionCancelled:
                add_job_log(job_id, f"Cancelled during: {video_path.name}", "warning")
                break
            except Exception as e:
                with _jobs_lock:
                    jobs[job_id].failed.append(
                        {
                            "path": str(video_path),
                            "error": str(e),
                        }
                    )
                add_job_log(job_id, f"Failed: {video_path.name} - {str(e)}", "error")

        # Final status update
        with _jobs_lock:
            job = jobs.get(job_id)
            if job:
                job.processed_videos = len(job.successful) + len(job.failed)
                job.status = "completed" if job.status != "cancelled" else "cancelled"
                job.completed_at = datetime.now().isoformat()
                final_status = job.status
                successful_count = len(job.successful)
                failed_count = len(job.failed)

        if final_status == "cancelled":
            add_job_log(
                job_id,
                f"Pipeline cancelled: {successful_count} completed, {failed_count} failed",
                "warning",
            )
        else:
            add_job_log(
                job_id,
                f"Pipeline finished: {successful_count} successful, {failed_count} failed",
                "info",
            )

    except ExtractionCancelled:
        _update_job(job_id, status="cancelled", completed_at=datetime.now().isoformat())
        add_job_log(job_id, "Pipeline cancelled by user", "warning")
        cleanup_gpu_memory()
        add_job_log(job_id, "GPU memory cleanup completed", "info")
    except Exception as e:
        with _jobs_lock:
            if job_id in jobs:
                jobs[job_id].status = "failed"
                jobs[job_id].failed.append({"path": "pipeline", "error": str(e)})
                jobs[job_id].completed_at = datetime.now().isoformat()
        add_job_log(job_id, f"Pipeline error: {str(e)}", "error")
        cleanup_gpu_memory()
        add_job_log(job_id, "GPU memory cleanup completed", "info")


def _run_mediapipe_extraction(
    video_path: Path,
    output_path: Path,
    config: ExtractionConfig,
    job_id: str,
) -> tuple[bool, str | None]:
    """Run MediaPipe extraction in a separate thread. Returns (success, error_message)."""
    try:
        from sltk.extraction.mediapipe import MediaPipeConfig, MediaPipeExtractor

        mp_config = MediaPipeConfig(
            device=config.device,
            enable_hands=config.mediapipe_hands,
            enable_face=config.mediapipe_face,
            compress=config.compress,
        )
        extractor = MediaPipeExtractor(mp_config)
        try:
            extractor.extract_from_video(video_path, output_path)
            add_job_log(job_id, f"  -> MediaPipe output: {output_path}", "info")
            return True, None
        finally:
            extractor.close()
    except Exception as e:
        add_job_log(job_id, f"  -> MediaPipe error: {e}", "error")
        return False, str(e)


def _run_wilor_extraction(
    video_path: Path,
    output_path: Path,
    config: ExtractionConfig,
    job_id: str,
) -> tuple[bool, str | None]:
    """Run WiLoR extraction in a separate thread. Returns (success, error_message)."""
    try:
        from sltk.extraction.wilor import WiLoRConfig, WiLoRExtractor

        wilor_config = WiLoRConfig(
            device=config.device,
            checkpoint_path=config.wilor_checkpoint,
            detector_path=config.wilor_detector,
            compress=config.compress,
            img_batch_size=config.wilor_img_batch_size,
            det_batch_size=config.wilor_det_batch_size,
        )
        extractor = WiLoRExtractor(wilor_config)
        try:
            result = extractor.extract_from_video(video_path, output_path)

            # Check if extraction actually succeeded
            if result.errors:
                error_msg = "; ".join(result.errors)
                raise RuntimeError(f"WiLoR extraction failed: {error_msg}")

            # Verify file was created
            if not output_path.exists():
                raise RuntimeError(f"WiLoR output file not created: {output_path}")

            add_job_log(job_id, f"  -> WiLoR output: {output_path}", "info")
            return True, None
        finally:
            extractor.close()
    except ImportError as e:
        add_job_log(job_id, f"  -> WiLoR not available: {e}", "warning")
        return False, f"Not available: {e}"
    except Exception as e:
        add_job_log(job_id, f"  -> WiLoR error: {e}", "error")
        return False, str(e)


async def process_single_video(
    video_path: Path,
    output_root: Path,
    config: ExtractionConfig,
    job: ExtractionJob,
    job_id: str = "",
):
    """Process a single video through the extraction pipeline."""
    video_stem = video_path.stem

    poses_dir = output_root / "poses"
    features_dir = output_root / "features"
    segments_dir = output_root / "segments"

    # Check for cancellation before starting
    raise_if_cancelled(job_id)

    # Get video info for progress tracking
    job.video_start_time = datetime.now().isoformat()
    job.current_frame = 0
    job.total_frames = 0
    job.stage_errors = {}

    # Determine which stages will run
    active_stages = []
    if config.enable_mediapipe:
        active_stages.append("mediapipe")
    if config.enable_nlf:
        active_stages.append("nlf")
    if config.enable_wilor:
        active_stages.append("wilor")
    active_stages.append("features")  # Always run if prerequisites exist
    if config.enable_segmentation:
        active_stages.append("segmentation")

    job.stages = active_stages
    job.stage_index = 0

    # Initialize stage statuses
    job.stage_statuses = {
        stage: StageStatus(name=stage, status="pending") for stage in active_stages
    }

    try:
        from sltk.extraction.base import get_video_info as get_vid_info

        video_info = get_vid_info(video_path)
        job.total_frames = video_info.get("num_frames", 0)
        add_job_log(
            job_id,
            f"  -> Video info: {job.total_frames} frames at {video_info.get('fps', 25):.1f} fps",
            "info",
        )
    except Exception as e:
        add_job_log(job_id, f"  -> Could not get video info: {e}", "warning")

    # Define output paths
    mediapipe_path = poses_dir / f"{video_stem}_mediapipe.h5"
    wilor_path = poses_dir / f"{video_stem}_wilor.h5"
    nlf_path = poses_dir / f"{video_stem}_nlf.h5"

    # Check which extractors need to run
    run_mediapipe = config.enable_mediapipe and (
        not config.skip_existing or not mediapipe_path.exists()
    )
    run_wilor = config.enable_wilor and (not config.skip_existing or not wilor_path.exists())

    # =========================================================================
    # Sequential extraction: Run MediaPipe then WiLoR to avoid memory issues
    # Running concurrently caused OOM errors due to both loading video frames
    # =========================================================================

    # Stage: MediaPipe extraction
    if run_mediapipe:
        stage_name = "mediapipe"
        job.current_stage = "MediaPipe extraction"
        job.stage_index = job.stages.index(stage_name) if stage_name in job.stages else 0

        if stage_name in job.stage_statuses:
            job.stage_statuses[stage_name].status = "running"
            job.stage_statuses[stage_name].start_time = datetime.now().isoformat()

        add_job_log(job_id, "  -> Running MediaPipe extraction...", "info")
        mp_success, mp_error = _run_mediapipe_extraction(video_path, mediapipe_path, config, job_id)

        if mp_success:
            if stage_name in job.stage_statuses:
                job.stage_statuses[stage_name].status = "completed"
        else:
            job.stage_errors[stage_name] = mp_error or "Unknown error"
            if stage_name in job.stage_statuses:
                job.stage_statuses[stage_name].status = "failed"
                job.stage_statuses[stage_name].error = mp_error

        if stage_name in job.stage_statuses:
            job.stage_statuses[stage_name].end_time = datetime.now().isoformat()

    elif config.enable_mediapipe and not run_mediapipe:
        # MediaPipe was skipped because output exists
        add_job_log(job_id, f"  -> Skipping MediaPipe (exists): {mediapipe_path.name}", "info")
        if "mediapipe" in job.stage_statuses:
            job.stage_statuses["mediapipe"].status = "skipped"

    # Stage: WiLoR extraction
    if run_wilor:
        stage_name = "wilor"
        job.current_stage = "WiLoR extraction"
        job.stage_index = job.stages.index(stage_name) if stage_name in job.stages else 1

        if stage_name in job.stage_statuses:
            job.stage_statuses[stage_name].status = "running"
            job.stage_statuses[stage_name].start_time = datetime.now().isoformat()

        add_job_log(job_id, "  -> Running WiLoR extraction...", "info")
        wilor_success, wilor_error = _run_wilor_extraction(video_path, wilor_path, config, job_id)

        if wilor_success:
            if stage_name in job.stage_statuses:
                job.stage_statuses[stage_name].status = "completed"
        else:
            job.stage_errors[stage_name] = wilor_error or "Unknown error"
            if stage_name in job.stage_statuses:
                job.stage_statuses[stage_name].status = "failed"
                job.stage_statuses[stage_name].error = wilor_error

        if stage_name in job.stage_statuses:
            job.stage_statuses[stage_name].end_time = datetime.now().isoformat()

    elif config.enable_wilor and not run_wilor:
        # WiLoR was skipped because output exists
        add_job_log(job_id, f"  -> Skipping WiLoR (exists): {wilor_path.name}", "info")
        if "wilor" in job.stage_statuses:
            job.stage_statuses["wilor"].status = "skipped"

    # Check for cancellation between stages
    raise_if_cancelled(job_id)

    # Stage: NLF extraction (full-body SMPL-X) - Always runs sequentially (GPU intensive)
    if config.enable_nlf:
        stage_name = "nlf"
        job.current_stage = "NLF extraction"
        job.stage_index = job.stages.index(stage_name) if stage_name in job.stages else 1

        if stage_name in job.stage_statuses:
            job.stage_statuses[stage_name].status = "running"
            job.stage_statuses[stage_name].start_time = datetime.now().isoformat()

        if not config.skip_existing or not nlf_path.exists():
            add_job_log(job_id, "  -> Running NLF extraction...", "info")
            try:
                from sltk.extraction.nlf import NLFConfig, NLFExtractor

                nlf_config = NLFConfig(
                    device=config.device,
                    model_path=config.nlf_checkpoint,
                    smplx_path=config.nlf_smplx_path,
                    compress=config.compress,
                )
                extractor = NLFExtractor(nlf_config)
                try:
                    result = extractor.extract_from_video(video_path, nlf_path)

                    # Check if extraction actually succeeded
                    if result.errors:
                        error_msg = "; ".join(result.errors)
                        raise RuntimeError(f"NLF extraction failed: {error_msg}")

                    # Verify file was created
                    if not nlf_path.exists():
                        raise RuntimeError(f"NLF output file not created: {nlf_path}")

                    add_job_log(job_id, f"  -> NLF output: {nlf_path}", "info")
                    if stage_name in job.stage_statuses:
                        job.stage_statuses[stage_name].status = "completed"
                finally:
                    extractor.close()
            except ImportError as e:
                job.stage_errors[stage_name] = f"Not available: {e}"
                if stage_name in job.stage_statuses:
                    job.stage_statuses[stage_name].status = "failed"
                    job.stage_statuses[stage_name].error = str(e)
                add_job_log(job_id, f"  -> NLF not available: {e}", "warning")
            except Exception as e:
                job.stage_errors[stage_name] = str(e)
                if stage_name in job.stage_statuses:
                    job.stage_statuses[stage_name].status = "failed"
                    job.stage_statuses[stage_name].error = str(e)
                add_job_log(job_id, f"  -> NLF error: {e}", "error")
        else:
            add_job_log(job_id, f"  -> Skipping NLF (exists): {nlf_path.name}", "info")
            if stage_name in job.stage_statuses:
                job.stage_statuses[stage_name].status = "skipped"

        if stage_name in job.stage_statuses:
            job.stage_statuses[stage_name].end_time = datetime.now().isoformat()

    # Check for cancellation between stages
    raise_if_cancelled(job_id)

    # Stage: Feature computation (angles, HaMeR)
    stage_name = "features"
    job.current_stage = "Feature computation"
    job.stage_index = job.stages.index(stage_name) if stage_name in job.stages else 2

    if stage_name in job.stage_statuses:
        job.stage_statuses[stage_name].status = "running"
        job.stage_statuses[stage_name].start_time = datetime.now().isoformat()

    angles_dir = features_dir / "angles"
    hamer_dir = features_dir / "hamer"
    angles_dir.mkdir(exist_ok=True)
    hamer_dir.mkdir(exist_ok=True)

    angles_path = angles_dir / f"{video_stem}.pt"
    hamer_path = hamer_dir / f"{video_stem}.lzma"

    # Check if we have both MediaPipe and WiLoR outputs
    mediapipe_path = poses_dir / f"{video_stem}_mediapipe.h5"
    wilor_path = poses_dir / f"{video_stem}_wilor.h5"

    if mediapipe_path.exists() and wilor_path.exists():
        if not config.skip_existing or not (angles_path.exists() and hamer_path.exists()):
            add_job_log(job_id, "  -> Computing angle and HaMeR features...", "info")
            try:
                from sltk.processing.features import load_features_from_h5, save_features

                angle_features, hamer_features = load_features_from_h5(
                    mediapipe_path=mediapipe_path,
                    wilor_path=wilor_path,
                )

                save_features(
                    angle_features=angle_features,
                    hamer_features=hamer_features,
                    angles_path=angles_path,
                    hamer_path=hamer_path,
                )

                feature_msg = (
                    f"  -> Features: angles={angle_features.shape}, "
                    f"hamer={hamer_features.shape}"
                )
                add_job_log(job_id, feature_msg, "info")
                if stage_name in job.stage_statuses:
                    job.stage_statuses[stage_name].status = "completed"
            except Exception as e:
                job.stage_errors[stage_name] = str(e)
                if stage_name in job.stage_statuses:
                    job.stage_statuses[stage_name].status = "failed"
                    job.stage_statuses[stage_name].error = str(e)
                add_job_log(job_id, f"  -> Feature computation error: {e}", "warning")
        else:
            add_job_log(
                job_id,
                f"  -> Skipping features (exist): {angles_path.name}, {hamer_path.name}",
                "info",
            )
            if stage_name in job.stage_statuses:
                job.stage_statuses[stage_name].status = "skipped"
    else:
        missing = []
        if not mediapipe_path.exists():
            missing.append("MediaPipe")
        if not wilor_path.exists():
            missing.append("WiLoR")
        job.stage_errors[stage_name] = f"Missing prerequisites: {', '.join(missing)}"
        if stage_name in job.stage_statuses:
            job.stage_statuses[stage_name].status = "skipped"
            job.stage_statuses[stage_name].error = f"Missing: {', '.join(missing)}"
        add_job_log(job_id, f"  -> Skipping features (missing: {', '.join(missing)})", "warning")

    if stage_name in job.stage_statuses:
        job.stage_statuses[stage_name].end_time = datetime.now().isoformat()

    # Check for cancellation between stages
    raise_if_cancelled(job_id)

    # Stage 4: Segmentation (v2 — uses WiLoR H5 directly)
    if config.enable_segmentation:
        stage_name = "segmentation"
        job.current_stage = "Segmentation"
        job.stage_index = job.stages.index(stage_name) if stage_name in job.stages else 3

        if stage_name in job.stage_statuses:
            job.stage_statuses[stage_name].status = "running"
            job.stage_statuses[stage_name].start_time = datetime.now().isoformat()

        segment_path = segments_dir / f"{video_stem}.eaf"

        if wilor_path.exists():
            if not config.skip_existing or not segment_path.exists():
                add_job_log(job_id, "  -> Running sign segmentation (v2)...", "info")
                try:
                    from sltk.segmentation import extract_segments, get_runner
                    from sltk.segmentation.h5_loader import h5_to_features
                    from sltk.segmentation.output import save_as_elan

                    runner = get_runner()
                    features = h5_to_features(wilor_path)
                    labels = runner.predict(features, include_begin=True)

                    # Save as ELAN file
                    save_as_elan({video_stem: labels}, segment_path, fps=25.0)

                    segments = extract_segments(labels)
                    seg_msg = f"  -> Segmented: {len(segments)} segments -> {segment_path.name}"
                    add_job_log(job_id, seg_msg, "info")
                    if stage_name in job.stage_statuses:
                        job.stage_statuses[stage_name].status = "completed"
                except ImportError as e:
                    job.stage_errors[stage_name] = f"Not available: {e}"
                    if stage_name in job.stage_statuses:
                        job.stage_statuses[stage_name].status = "failed"
                        job.stage_statuses[stage_name].error = str(e)
                    add_job_log(job_id, f"  -> Segmentation not available: {e}", "warning")
                except Exception as e:
                    job.stage_errors[stage_name] = str(e)
                    if stage_name in job.stage_statuses:
                        job.stage_statuses[stage_name].status = "failed"
                        job.stage_statuses[stage_name].error = str(e)
                    add_job_log(job_id, f"  -> Segmentation error: {e}", "error")
            else:
                add_job_log(
                    job_id, f"  -> Skipping segmentation (exists): {segment_path.name}", "info"
                )
                if stage_name in job.stage_statuses:
                    job.stage_statuses[stage_name].status = "skipped"
        else:
            error_msg = "WiLoR H5 not available for segmentation"
            job.stage_errors[stage_name] = error_msg
            if stage_name in job.stage_statuses:
                job.stage_statuses[stage_name].status = "skipped"
                job.stage_statuses[stage_name].error = error_msg
            add_job_log(job_id, "  -> Segmentation skipped: WiLoR H5 not available", "warning")

        if stage_name in job.stage_statuses:
            job.stage_statuses[stage_name].end_time = datetime.now().isoformat()


# ============================================================================
# Export for admin router
# ============================================================================


def get_job_stats() -> dict:
    """Get job statistics for admin endpoints."""
    with _jobs_lock:
        total_jobs = len(jobs)
        status_counts: dict[str, int] = {}
        for job in jobs.values():
            status_counts[job.status] = status_counts.get(job.status, 0) + 1

    with _logs_lock:
        total_log_entries = sum(len(logs) for logs in job_logs.values())

    return {
        "total_jobs": total_jobs,
        "status_counts": status_counts,
        "total_log_entries": total_log_entries,
        "config": {
            "ttl_hours": JOB_TTL_HOURS,
            "max_jobs": MAX_JOBS,
            "cleanup_interval_seconds": JOB_CLEANUP_INTERVAL_SECONDS,
        },
    }



"""Post-processing functions for BIO label predictions.

Contains label correction (label_fixes), segment extraction, and BEGIN tag merging.

Adapted from segmenterv2/segmenter/inference/postprocess.py.
"""

from __future__ import annotations

import torch


def label_fixes(
    labels: torch.Tensor,
    min_segment_len: int = 2,
    min_start_len: int = 6,
) -> torch.Tensor:
    """Two-pass correction of BIO predictions.

    Pass 1 (reverse): Insert BEGIN at the start of IN-runs >= min_segment_len.
    Pass 2 (forward): Collapse consecutive BEGINs (keep first, rest become IN).

    Args:
        labels: 1D tensor of BIO predictions (0=OUT, 1=IN, 2=BEGIN).
        min_segment_len: Minimum IN-run length to trigger BEGIN insertion.
        min_start_len: Minimum IN-run length at sequence start to insert BEGIN.

    Returns:
        Corrected labels tensor (new tensor, input is not mutated).
    """
    labels = labels.clone()
    length = len(labels)
    if length == 0:
        return labels

    # Pass 1: Reverse pass to insert BEGIN at start of IN-runs
    one_counter = 0
    for index in reversed(range(length)):
        if labels[index] == 1:
            one_counter += 1
        else:
            if one_counter >= min_segment_len:
                if index + 1 < length and labels[index + 1] != 2:
                    labels[index + 1] = 2
            one_counter = 0

    # Edge case: IN-run starting from index 0
    if one_counter >= min_start_len and labels[0] != 2:
        labels[0] = 2

    # Pass 2: Forward pass to collapse consecutive BEGINs
    i = 0
    while i < length:
        if labels[i] == 2:
            j = i + 1
            while j < length and labels[j] == 2:
                labels[j] = 1
                j += 1
            i = j
        else:
            i += 1

    return labels


def extract_segments(prediction: torch.Tensor) -> list[tuple[int, int]]:
    """Extract (start_frame, end_frame) tuples from BIO predictions.

    A segment starts at BEGIN or at the first IN after OUT.
    A segment ends when OUT is encountered or a new BEGIN starts.

    Args:
        prediction: 1D tensor of BIO predictions (0=OUT, 1=IN, 2=BEGIN).

    Returns:
        List of (start_frame, end_frame) tuples (inclusive).
    """
    pred = prediction.clone()
    for i in reversed(range(1, len(pred))):
        if pred[i] == 2 and pred[i - 1] == 2:
            pred[i] = 1

    segments = []
    start_time = None
    for i, p in enumerate(pred):
        if p == 1 or p == 2:
            if start_time is not None and p == 2:
                segments.append((start_time, i - 1))
                start_time = None
            if start_time is None:
                start_time = i
        elif p == 0:
            if start_time is not None:
                segments.append((start_time, i - 1))
                start_time = None

    if start_time is not None:
        segments.append((start_time, len(pred) - 1))

    return segments


def merge_begin_tags(labels: torch.Tensor) -> torch.Tensor:
    """Merge BEGIN tags into IN tags. BEGIN(2) -> IN(1).

    Args:
        labels: 1D tensor of BIO predictions.

    Returns:
        New tensor with all BEGIN (2) values replaced by IN (1).
    """
    result = labels.clone()
    result[result == 2] = 1
    return result



"""
Utility functions for SLTK.

Research utilities for dataset analysis, batch processing, and ML workflows.
"""

from sltk.utils.research import (
    DatasetStats,
    align_poses_to_segments,
    batch_extract_poses,
    compare_vocabularies,
    compute_dataset_statistics,
    compute_pose_coverage,
    create_vocabulary_mapping,
    export_segments_for_training,
    find_gloss_examples,
)

__all__ = [
    "DatasetStats",
    "compute_dataset_statistics",
    "compare_vocabularies",
    "find_gloss_examples",
    "export_segments_for_training",
    "create_vocabulary_mapping",
    "batch_extract_poses",
    "compute_pose_coverage",
    "align_poses_to_segments",
]



"""Programmatic API for NMS detection, extracted from cli.py logic."""

from __future__ import annotations

import logging
from pathlib import Path

from .constants import ALL_DETECTORS, THRESHOLD_DEFAULTS
from .detectors.blinks import detect_blinks
from .detectors.pipeline import detect_all_nms
from .io.exporters import export_csv, export_elan, export_json
from .io.loaders import load_smpl_eye_data, load_teaser_data, resolve_smpl_path
from .models import BlinkEvent, NMSEvent, TrackingQuality
from .processing.adaptive import compute_adaptive_thresholds
from .processing.validity import compute_validity_mask, segment_valid_regions

logger = logging.getLogger(__name__)


def detect_nms(
    h5_path: str,
    *,
    person: str = "0",
    detectors: set[str] | list[str] | None = None,
    no_adaptive: bool = False,
    no_gaze: bool = False,
    smpl_path: str | None = None,
    smpl_root: str | None = None,
    video_path: str | None = None,
    # Blink params
    prominence: float | None = None,
    sigma: float = 1.5,
    min_distance: int = 5,
    min_width: int = 1,
    max_width: int = 18,
    min_score: float = 0.50,
    min_segment: int = 2,
    yaw_limit: float = 0.8,
    # NMS thresholds
    thresholds: dict[str, float] | None = None,
) -> tuple[list[BlinkEvent], list[NMSEvent], TrackingQuality]:
    """Run NMS detection on a TEASER H5 file.

    This is the main programmatic entry point, extracting the core detection
    pipeline from the CLI without argparse overhead.

    Args:
        h5_path: Path to TEASER HDF5 file.
        person: Person group in HDF5 file (default "0").
        detectors: Set of detector names to run, or None for blink-only.
            Use {"all"} or ALL_DETECTORS for everything.
            Valid: blink, nod, shake, tilt, mouth, eyebrow, gaze, squint, lips.
        no_adaptive: Disable adaptive thresholds.
        no_gaze: Skip eye gaze detection.
        smpl_path: Explicit path to SMPL NLF file for gaze.
        smpl_root: Root directory to search for SMPL files.
        video_path: Path to source video (for ELAN export).
        prominence: Blink detection prominence threshold.
        sigma: Gaussian smoothing sigma.
        min_distance: Min frames between blinks.
        min_width: Min blink width in frames.
        max_width: Max blink width in frames.
        min_score: Min detection confidence for frame validity.
        min_segment: Min valid segment length in frames.
        yaw_limit: Max head yaw in radians for valid frames.
        thresholds: Dict of threshold overrides (keys from THRESHOLD_DEFAULTS).

    Returns:
        Tuple of (blinks, nms_events, quality).
    """
    if not Path(h5_path).exists():
        raise FileNotFoundError(f"HDF5 file not found: {h5_path}")

    # Parse detectors
    if detectors is not None:
        det_set = set(detectors)
        if "all" in det_set:
            det_set = ALL_DETECTORS.copy()
            run_blinks = True
        else:
            run_blinks = "blink" in det_set
            det_set.discard("blink")
            det_set &= ALL_DETECTORS
        nms_detectors = det_set if det_set else None
        run_nms = nms_detectors is not None
    else:
        run_blinks = True
        nms_detectors = None
        run_nms = False

    # Load data
    eyelid, successful_frames, boxes, pose, scores, expression, jaw_pose, quality = (
        load_teaser_data(
            h5_path,
            person,
            load_expression=True,
            load_nms=run_nms,
        )
    )

    skip_blinks = False
    if quality.active_eye == -1:
        if not run_nms:
            return [], [], quality
        skip_blinks = True

    # Load SMPL eye gaze data
    leye_pose, reye_pose = None, None
    if run_nms and not no_gaze:
        resolved_smpl = None
        if smpl_path:
            p = Path(smpl_path)
            if p.exists():
                resolved_smpl = p
        if resolved_smpl is None:
            root = Path(smpl_root) if smpl_root else None
            resolved_smpl = resolve_smpl_path(h5_path, root) if root else resolve_smpl_path(h5_path)
        if resolved_smpl is not None:
            leye_pose, reye_pose = load_smpl_eye_data(
                resolved_smpl,
                successful_frames,
                boxes,
            )

    # Validity masks
    valid = compute_validity_mask(
        eyelid,
        scores,
        pose,
        min_score=min_score,
        yaw_limit=yaw_limit,
        require_eyelid=True,
    )

    if run_nms:
        nms_valid = compute_validity_mask(
            eyelid,
            scores,
            pose,
            min_score=max(0.30, min_score - 0.20),
            yaw_limit=yaw_limit,
            require_eyelid=False,
        )
    else:
        nms_valid = valid

    # Segment valid regions
    segments = segment_valid_regions(valid, min_segment=min_segment)
    nms_segments = (
        segment_valid_regions(nms_valid, min_segment=min_segment) if run_nms else segments
    )

    if not segments and not (run_nms and nms_segments):
        return [], [], quality

    # Adaptive thresholds
    user_thresholds = thresholds or {}

    if no_adaptive:
        adaptive = {}
    else:
        adaptive = compute_adaptive_thresholds(
            pose,
            jaw_pose,
            expression,
            nms_valid if run_nms else valid,
            eyelid=eyelid,
        )

    def _pick(user_val, key):
        """Use user override > adaptive > default."""
        if key in user_thresholds:
            return user_thresholds[key]
        if user_val is not None:
            return user_val
        return adaptive.get(key, THRESHOLD_DEFAULTS.get(key, 0.0))

    blink_prom = _pick(prominence, "prominence")

    # Detect blinks
    blinks = []
    if run_blinks and not skip_blinks and segments:
        blinks = detect_blinks(
            eyelid,
            successful_frames,
            valid,
            segments,
            sigma=sigma,
            prominence=blink_prom,
            min_distance=min_distance,
            min_width=min_width,
            max_width=max_width,
            expression=expression,
        )

    # Detect NMS events
    nms_events = []
    if run_nms and nms_detectors:
        nms_events = detect_all_nms(
            pose,
            jaw_pose,
            expression,
            successful_frames,
            nms_valid,
            nms_segments,
            detectors=nms_detectors,
            nod_amplitude=_pick(None, "nod_amplitude"),
            shake_amplitude=_pick(None, "shake_amplitude"),
            tilt_threshold=_pick(None, "tilt_threshold"),
            mouth_threshold=_pick(None, "mouth_threshold"),
            eyebrow_threshold=_pick(None, "eyebrow_threshold"),
            leye_pose=leye_pose,
            reye_pose=reye_pose,
            gaze_threshold=user_thresholds.get("gaze_threshold", 1.5),
            eyelid=eyelid,
            blink_events=blinks,
            squint_threshold=_pick(None, "squint_threshold"),
            lip_threshold=_pick(None, "lip_threshold"),
        )

    return blinks, nms_events, quality


def export_results(
    blinks: list[BlinkEvent],
    nms_events: list[NMSEvent],
    h5_path: str,
    *,
    output_dir: str | None = None,
    formats: list[str] | None = None,
    participant_id: str | None = None,
    media_path: str = "",
) -> dict[str, str]:
    """Export detection results to the requested formats.

    Args:
        blinks: Detected blink events.
        nms_events: Detected NMS events.
        h5_path: Original H5 file path (for naming).
        output_dir: Output directory (default: same as h5_path).
        formats: List of formats to export: "elan", "csv", "json".
        participant_id: Participant ID for ELAN tier names.
        media_path: Media file path for ELAN export.

    Returns:
        Dict of format -> output path.
    """
    if not formats:
        return {}

    stem = Path(h5_path).stem
    base_dir = Path(output_dir) if output_dir else Path(h5_path).parent
    base_dir.mkdir(parents=True, exist_ok=True)

    export_paths = {}

    for fmt in formats:
        fmt = fmt.lower().strip()
        if fmt == "elan":
            path = str(base_dir / f"{stem}_nms.eaf")
            export_elan(
                blinks,
                path,
                media_path=media_path,
                nms_events=nms_events if nms_events else None,
                participant_id=participant_id,
                h5_path=h5_path,
            )
            export_paths["elan"] = path
        elif fmt == "csv":
            path = str(base_dir / f"{stem}_nms.csv")
            export_csv(blinks, path, nms_events=nms_events if nms_events else None)
            export_paths["csv"] = path
        elif fmt == "json":
            path = str(base_dir / f"{stem}_nms.json")
            export_json(blinks, path, nms_events=nms_events if nms_events else None)
            export_paths["json"] = path

    return export_paths
